"""
Two examples for arbitrary finite domains:

    - one manually coded
    - one with arbitrary (unnumbered) sys TS states
"""
# will probably need also a flag which type to use

from tulip import spec, synth
from tulip import transys as trs

env_vars = {'park'}
env_prog = '!park'

sys_vars = {}
sys_vars['loc'] = ['a', 'b', 'c', 'd', 'e', 'foo']

sys_init = {'loc= "a"'}
sys_safe = {
    'loc="a" -> next(loc="b" || loc="d")',
    'loc="b" -> next(loc="a" || loc="e" || loc="c")',
    'loc="c" -> next(loc="b" || loc="foo")',
    'loc="d" -> next(loc="a" || loc="e")',
    'loc="e" -> next(loc="d" || loc="b" || loc="foo")',
    'loc="foo" -> next(loc="e" || loc="c")',
}

sys_vars['mem'] = 'boolean'
sys_init |= {'mem'}
sys_safe |= {'next(mem) <-> (loc="a") || (mem && !park)'}
sys_prog = {'mem'}

specs = spec.GRSpec(env_vars=env_vars, sys_vars=sys_vars,
                    sys_init=sys_init, sys_safety=sys_safe,
                    env_prog=env_prog, sys_prog=sys_prog)
specs.moore = False
specs.qinit = '\A \E'
ctrl = synth.synthesize('omega', specs)
ctrl.save('gr1_arbitrary_set0.pdf')

"""
2nd example
"""
sys = trs.FTS()

states = {'a', 'b', 'c', 'd', 'e', 'foo'}
sys.states.add_from(states)
sys.states.initial.add('a')

sys.transitions.add_from(
    [('a', x) for x in {'b', 'd'}] +
    [('b', x) for x in {'a', 'e', 'c'}] +
    [('c', x) for x in {'b', 'foo'}] +
    [('d', x) for x in {'a', 'e'}] +
    [('e', x) for x in {'d', 'b', 'foo'}] +
    [('foo', x) for x in {'e', 'c'}]
)

sys.atomic_propositions |= {'cave'}
sys.states.add('a', ap={'cave'})

sys_vars = {'mem': 'boolean'}
sys_init = {'mem'}

# if we don't want to use an extra AP to label a,
# then we need to either know that tulip internally names
# the state variable as 'loc', or
# provide an option to pass a name for the state var.
# The latter approach can help a user avoid name
# conflicts, in case, e.g., she wants to reserve
# 'loc' for different use
sys_safe = {'next(mem) <-> (cave) || (mem && !park)'}

specs = spec.GRSpec(env_vars=env_vars, sys_vars=sys_vars,
                    sys_init=sys_init, sys_safety=sys_safe,
                    env_prog=env_prog, sys_prog=sys_prog)
specs.moore = False
specs.qinit = '\A \E'
ctrl = synth.synthesize('omega', specs, sys=sys)
ctrl.save('gr1_arbitrary_set1.pdf')



# Copyright (c) 2013-2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Automata Module"""
from __future__ import absolute_import
import logging
import copy
from collections import Iterable
from pprint import pformat
from tulip.transys.labeled_graphs import (
    LabeledDiGraph, str2singleton, prepend_with)
from tulip.transys.mathset import SubSet, PowerSet
from tulip.transys.transys import GameGraph


logger = logging.getLogger(__name__)
_hl = 40 * '-'


class FiniteStateAutomaton(LabeledDiGraph):
    """Set of sequences described with a graph and a condition.

    It has:
        - states
        - states.initial
        - states.accepting (types have names, and classes)
        - alphabet = set of symbols that label edges.


    Note
    ====
    If all paths in the graph belong to the set you
    want to describe, then just use L{FiniteTransitionSystem}.

    To describe an input-output function (which is a set too),
    it is more convenient to use L{FiniteStateMachine}.


    See Also
    ========
    L{BA}, L{RabinAutomaton}.
    """

    def __init__(
            self, deterministic=False,
            accepting_states_type=None,
            atomic_proposition_based=True,
            symbolic=False,
    ):
        """Initialize FiniteStateAutomaton.

        Additional keyword arguments are passed to L{LabeledDiGraph.__init__}.

        @param atomic_proposition_based: If False, then the alphabet
            is represented by a set.  If True, then the alphabet is
            represented by a powerset 2^AP.
        """
        self.atomic_proposition_based = atomic_proposition_based
        self.symbolic = symbolic

        # edge labeling
        if symbolic:
            alphabet = None  # no checks
        else:
            if atomic_proposition_based:
                alphabet = PowerSet([])
                self.atomic_propositions = alphabet.math_set
            else:
                alphabet = set()
        self.alphabet = alphabet

        edge_label_types = [
            {'name': 'letter',
             'values': alphabet,
             'setter': True}]
        super(FiniteStateAutomaton, self).__init__(
            edge_label_types=edge_label_types)
        # accepting states
        if accepting_states_type is None:
            self._accepting = SubSet(self.states)
            self._accepting_type = SubSet
        else:
            self._accepting = accepting_states_type(self)
            self._accepting_type = accepting_states_type
        self.states.accepting = self._accepting
        # used before label value
        self._transition_dot_label_format = {'letter': '',
                                             'type?label': '',
                                             'separator': r'\\n'}
        self._transition_dot_mask = dict()

        self.dot_node_shape = {'normal': 'circle',
                               'accepting': 'doublecircle'}
        self.default_export_fname = 'fsa'
        self.automaton_type = 'Finite State Automaton'

    @property
    def accepting(self):
        return self._accepting

    def __str__(self):
        s = (
            _hl + '\n' + self.automaton_type + ': ' +
            self.name + '\n' + _hl + '\n' +
            'States:\n' +
            pformat(self.states(data=False), indent=3) + 2 * '\n' +
            'Initial States:\n' +
            pformat(self.states.initial, indent=3) + 2 * '\n' +
            'Accepting States:\n' +
            pformat(self.states.accepting, indent=3) + 2 * '\n')
        if self.atomic_proposition_based:
            s += 'Input Alphabet Letters (\in 2^AP):\n\t'
        else:
            if hasattr(self, 'alphabet'):
                s += ('Input Alphabet Letters:\n\t' +
                      str(self.alphabet) + 2 * '\n')
        s += (
            'Transitions & labeling w/ Input Letters:\n' +
            pformat(self.transitions(data=True), indent=3) +
            '\n' + _hl + '\n')
        return s

    def remove_node(self, node):
        """Remove state (also referred to as "node").

        More than a wrapper since the state is also removed from the
        accepting set if present.
        """
        # intercept to remove also from accepting states
        self.accepting.remove(node)
        super(FiniteStateAutomaton, self).remove_node(node)


class FiniteWordAutomaton(FiniteStateAutomaton):
    """Finite-word finite-state automaton.

    By default non-deterministic (NFA).
    To enforce determinism (DFA):

    >>> a = FiniteWordAutomaton(deterministic=True)
    """
    def __init__(self, deterministic=False,
                 atomic_proposition_based=True):
        super(FiniteWordAutomaton, self).__init__(
            deterministic=deterministic,
            atomic_proposition_based=atomic_proposition_based)
        self.automaton_type = 'Finite-Word Automaton'


def dfa2nfa(dfa):
    """Copy DFA to an NFA, so remove determinism restriction."""
    nfa = copy.deepcopy(dfa)
    nfa.transitions._deterministic = False
    nfa.automaton_type = 'Non-Deterministic Finite Automaton'
    return nfa


class OmegaAutomaton(FiniteStateAutomaton):
    def __init__(self, *args, **kwargs):
        super(OmegaAutomaton, self).__init__(*args, **kwargs)


class BuchiAutomaton(OmegaAutomaton):
    def __init__(
            self, deterministic=False,
            atomic_proposition_based=True,
            symbolic=False
    ):
        super(BuchiAutomaton, self).__init__(
            deterministic=deterministic,
            atomic_proposition_based=atomic_proposition_based,
            symbolic=symbolic)
        self.automaton_type = 'Buchi Automaton'


class BA(BuchiAutomaton):
    """Alias to L{BuchiAutomaton}."""

    def __init__(self, **args):
        super(BA, self).__init__(**args)


def tuple2ba(S, S0, Sa, Sigma_or_AP, trans, name='ba', prepend_str=None,
             atomic_proposition_based=True):
    """Create a Buchi Automaton from a tuple of fields.

    defines Buchi Automaton by a tuple (S, S0, Sa, \\Sigma, trans)
    (maybe replacing \\Sigma by AP since it is an AP-based BA ?)

    See Also
    ========
    L{tuple2fts}

    @param S: set of states
    @param S0: set of initial states, must be \\subset S
    @param Sa: set of accepting states
    @param Sigma_or_AP: Sigma = alphabet
    @param trans: transition relation, represented by list of triples::
            [(from_state, to_state, guard), ...]
    where guard \\in \\Sigma.

    @param name: used for file export
    @type name: str

    @rtype: L{BuchiAutomaton}
    """
    # args
    if not isinstance(S, Iterable):
        raise TypeError('States S must be iterable, even for single state.')
    if not isinstance(S0, Iterable) or isinstance(S0, str):
        S0 = [S0]
    if not isinstance(Sa, Iterable) or isinstance(Sa, str):
        Sa = [Sa]
    # comprehensive names
    states = S
    initial_states = S0
    accepting_states = Sa
    alphabet_or_ap = Sigma_or_AP
    transitions = trans
    # prepending states with given str
    if prepend_str:
        logger.debug('Given string:\n\t' + str(prepend_str) + '\n' +
                     'will be prepended to all states.')
    states = prepend_with(states, prepend_str)
    initial_states = prepend_with(initial_states, prepend_str)
    accepting_states = prepend_with(accepting_states, prepend_str)

    ba = BuchiAutomaton(atomic_proposition_based=atomic_proposition_based)
    ba.name = name

    ba.states.add_from(states)
    ba.states.initial |= initial_states
    ba.states.accepting |= accepting_states

    if atomic_proposition_based:
        ba.alphabet.math_set |= alphabet_or_ap
    else:
        ba.alphabet.add(alphabet_or_ap)
    for transition in transitions:
        (from_state, to_state, guard) = transition
        [from_state, to_state] = prepend_with([from_state, to_state],
                                              prepend_str)
        # convention
        if atomic_proposition_based:
            if guard is None:
                guard = set()
            guard = str2singleton(guard)
        ba.transitions.add(from_state, to_state, letter=guard)
    return ba


class RabinPairs(object):
    """Acceptance pairs for Rabin automaton.

    Each pair defines an acceptance condition.
    A pair (L, U) comprises of:
        - a set L of "good" states
        - a set U of "bad" states
    L,U must each be a subset of States.

    A run: (q0, q1, ...) is accepted if for at least one Rabin Pair,
    it in intersects L an inf number of times, but U only finitely.

    Internally a list of 2-tuples of SubSet objects is maintained::
        [(L1, U1), (L2, U2), ...]
    where: Li, Ui, are SubSet objects, with superset
    the Rabin automaton's States.

    Caution
    =======
    Here and in ltl2dstar documentation L denotes a "good" set.
    U{[BK08] <http://tulip-control.sourceforge.net/doc/bibliography.html#bk08>}
    denote the a "bad" set with L.  To avoid ambiguity, attributes:
    .good, .bad were used here.

    Example
    =======
    >>> dra = RabinAutomaton()
    >>> dra.states.add_from([1, 2, 3] )
    >>> dra.states.accepting.add([1], [2] )
    >>> dra.states.accepting

    >>> dra.states.accepting.good(1)

    >>> dra.states.accepting.bad(1)

    See Also
    ========
      - L{RabinAutomaton}
      - Def. 10.53, p.801, U{[BK08]
        <http://tulip-control.sourceforge.net/doc/bibliography.html#bk08>}
      - U{ltl2dstar<http://ltl2dstar.de/>} documentation
    """

    def __init__(self, automaton_states):
        self._states = automaton_states
        self._pairs = []

    def __str__(self):
        s = 'L = Good states, U = Bad states\n' + 30 * '-' + '\n'
        for index, (good, bad) in enumerate(self._pairs):
            s += (
                'Pair: ' + str(index) + ', L = ' + str(good) +
                ', U = ' + str(bad) + '\n')
        return s

    def __getitem__(self, index):
        return self._pairs[index]

    def __iter__(self):
        return iter(self._pairs)

    def __call__(self):
        """Get list of 2-tuples (L, U) of good-bad sets of states."""
        return list(self._pairs)

    def add(self, good_states, bad_states):
        """Add new acceptance pair (L, U).

        See Also
        ========
        remove, add_states, good, bad

        @param good_states: set L of good states for this pair
        @type good_states: container of valid states

        @param bad_states: set U of bad states for this pair
        @type bad_states: container of valid states
        """
        good_set = SubSet(self._states)
        good_set |= good_states
        bad_set = SubSet(self._states)
        bad_set |= bad_states
        self._pairs.append((good_set, bad_set))

    def remove(self, good_states, bad_states):
        """Delete pair (L, U) of good-bad sets of states.

        Note
        ====
        Removing a pair which is not last changes
        the indices of all other pairs, because internally
        a list is used.

        The sets L,U themselves (good-bad) are required
        for the deletion, instead of an index, to prevent
        acceidental deletion of an unintended pair.

        Get the intended pair using __getitem__ first
        (or in any other way) and them call remove.
        If the pair is corrent, then the removal will
        be successful.

        See Also
        ========
        add

        @param good_states: set of good states of this pair
        @type good_states: iterable container
        """
        good_set = SubSet(self._states)
        good_set |= good_states
        bad_set = SubSet(self._states)
        bad_set |= bad_states
        self._pairs.remove((good_set, bad_set))

    def add_states(self, pair_index, good_states, bad_states):
        try:
            self._pairs[pair_index][0].add_from(good_states)
            self._pairs[pair_index][1].add_from(bad_states)
        except IndexError:
            raise Exception("A pair with pair_index doesn't exist.\n" +
                            'Create a new one by callign .add.')

    def good(self, index):
        """Return set L of "good" states for this pair.

        @param index: number of Rabin acceptance pair
        @type index: int <= current total number of pairs
        """
        return self._pairs[index][0]

    def bad(self, index):
        """Return set U of "bad" states for this pair.

        @param index: number of Rabin acceptance pair
        @type index: int <= current total number of pairs
        """
        return self._pairs[index][1]

    def has_superset(self, superset):
        """Return true if the given argument is the superset."""
        return superset is self._states


class RabinAutomaton(OmegaAutomaton):
    """Rabin automaton.

    See Also
    ========
    L{DRA}, L{BuchiAutomaton}
    """

    def __init__(self, deterministic=False,
                 atomic_proposition_based=False):
        super(RabinAutomaton, self).__init__(
            deterministic=deterministic,
            accepting_states_type=RabinPairs,
            atomic_proposition_based=atomic_proposition_based)
        self.automaton_type = 'Rabin Automaton'


class DRA(RabinAutomaton):
    """Deterministic Rabin Automaton.

    See Also
    ========
    L{RabinAutomaton}
    """

    def __init__(self, atomic_proposition_based=True):
        super(DRA, self).__init__(
            deterministic=True,
            atomic_proposition_based=atomic_proposition_based)
        self.automaton_type = 'Deterministic Rabin Automaton'


class ParityGame(GameGraph):
    """GameGraph equipped with coloring.

    Define as C{k} the highest color that
    occurs infinitely many times.

    If C{k} is even, then Player 0 wins.
    Otherwise Player 1 wins (C{k} is odd).
    So the winner is Player (k mod 2).

    To define the number of colors C{c}:

    >>> p = ParityGame(c=4)

    Note that the colors are: 0, 1, ..., c-1

    See also
    ========
    L{transys.GameGraph}
    """

    def __init__(self, c=2):
        node_label_types = [{
            'name': 'color',
            'values': range(c),
            'default': 0}]
        super(ParityGame, self).__init__(node_label_types, [])

    def __str__(self):
        s = (
            'Parity Game\n'
            '-----------\n'
            'n: node, p: player, c: color\n\n')
        for node, attr in self.states(data=True):
            s += 'nd = {node}, p = {player}, c = {color}\n'.format(
                npde=node, player=attr['player'], color=attr['color'])
        s += '\n{t}'.format(t=self.transitions)
        return s

    @property
    def max_color(self):
        max_c = -1
        # node = None
        for x in self:
            if self.node[x]['color'] > max_c:
                max_c = self.node[x]['color']
                # node = x
        return max_c



# Copyright (c) 2013-2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Transition System Module"""
from __future__ import absolute_import
import logging
from collections import Iterable
from pprint import pformat
from tulip.transys.labeled_graphs import (
    LabeledDiGraph, str2singleton, prepend_with)
from tulip.transys.mathset import PowerSet, MathSet
# inline imports
#
# from tulip.transys.export import graph2promela


logger = logging.getLogger(__name__)
_hl = 40 * '-'


class KripkeStructure(LabeledDiGraph):
    """Directed graph with labeled vertices and initial vertices.

    References
    ==========
    1. Kripke S.
      Semantical Considerations on Modal Logic
      Acta Philosophica Fennica, 16, pp. 83-94, 1963

    2. Clarke E.M.; Grumberg O.; Peled D.A.
      Model Checking, MIT Press, 1999, p.14

    3. Schneider K.
      Verification of Reactive Systems
      Springer, 2004, Def. 2.1, p.45
    """

    def __init__(self):
        ap_labels = PowerSet()
        node_label_types = [
            {'name': 'ap',
             'values': ap_labels,
             'setter': ap_labels.math_set,
             'default': set()}]
        super(KripkeStructure, self).__init__(node_label_types)
        self.atomic_propositions = self.ap
        # dot formatting
        self._state_dot_label_format = {
            'ap': '',
            'type?label': '',
            'separator': r'\\n'}
        self.dot_node_shape = {'normal': 'rectangle'}
        self._state_dot_label_format = {
            'ap': '',
            'type?label': '',
            'separator': r'\\n'}
        self._transition_dot_label_format = {
            'type?label': ':',
            'separator': r'\\n'}
        self._transition_dot_mask = dict()

    def __str__(self):
        s = (
            'Kripke Structure: ' + self.name + '\n' +
            _hl + '\n' +
            'Atomic Propositions (APs):\n\t' +
            pformat(self.atomic_propositions, indent=3) + 2 * '\n' +
            'States labeled with sets of APs:\n' +
            _dumps_states(self) + 2 * '\n' +
            'Initial States:\n' +
            pformat(self.states.initial, indent=3) + 2 * '\n' +
            'Transitions:\n' +
            pformat(self.transitions(), indent=3) +
            '\n' + _hl + '\n')
        return s


class FiniteTransitionSystem(LabeledDiGraph):
    """Kripke structure with labeled states and edges.

    Who controls the state
    ======================
    To define who "moves the token" between vertices in
    the graph, set the attribute:

    >>> g = FiniteTransitionSystem()
    >>> g.owner = 'sys'

    This means that when there are more than one transition
    enabled, then the system picks the next state.

    The other option is:

    >>> g.owner = 'env'

    so the environment picks the next state.

    State labeling
    ==============
    The state labels are sets of atomic propositions,
    similar to a L{KripkeStructure}.

    In principle some of the propositions that label states
    could be controlled by either of the players,
    but this would lead to less straightforward semantics.

    You can achieve the same effect by using actions of
    the opponent.

    It is a matter of future experimentation whether
    this capability will be introduced, by partitioning
    the props into C{env_props} and C{sys_props}
    (similar to C{env_vars}, C{sys_vars} in L{GRSpec}).

    Edge labeling
    =============
    Edge labels are called "actions".

    The edge labeling is syntactic sugar for
    labels that are shifted to the target states of
    those edges. So edge labeling is not an essential
    difference from Kripke structures.

    Not to be confused with the term:
    "Labeled Transition System"
    found in the literature.

    Also, it differs from the definition in Baier-Katoen
    in that actions are not mere reading aid,
    but are interpreted as propositions as explained above.

    Besides, edge labeling usually allows for
    graphs with fewer vertices than the corresponding
    Kripke structure.

    Open vs Closed
    ==============
    The essential difference from Kripke structures
    is the partition of atomic propositions into
    input/output sets.

    If the set of inputs is empty, then the system is closed.
    Otherwise it is an open system.
    Open systems have an environment, closed don't.

    Alternatively, FTS can be thought of as a shorthand
    for defining a vertex-labeled game graph,
    or equivalently a game structure.

    System and environment actions
    ==============================
    The only significant difference is in transition labeling.
    For closed systems, each transition is labeled with a system action.
    So each transition label comprises of a single sublabel,
    the system action.

    For open systems, each transition is labeled with 2 sublabels:
        - The first sublabel is a system action,
        - the second an environment action.

    Mutual exclusion of actions
    ===========================
    Constraints on actions can be defined
    similarly to L{FTS} actions by setting the fields:

        - C{ofts.env_actions_must}
        - C{ofts.sys_actions_must}

    The default constraint is 'xor'.

    sys.sys_actions_must: select constraint on actions. Options:

        - C{'mutex'}: at most 1 action True each time
        - C{'xor'}: exactly 1 action True each time
        - C{'none'}: no constraint on action values

    The xor constraint can prevent the environment from
    blocking the system by setting all its actions to False.

    The action are taken when traversing an edge.
    Each edge is annotated by a single action.
    If an edge (s1, s2) can be taken on two transitions,
    then 2 copies of that same edge are stored.
    Each copy is annotated using a different action,
    the actions must belong to the same action set.
    That action set is defined as a set instance.
    This description is a (closed) L{FTS}.

    The system and environment actions are associated with an edge
    of a reactive system. To store these, mutliple labels are used
    and their sets are encapsulated within the same C{FTS}.

    Example
    =======
    In the following C{None} represents the empty set, subset of AP.
    First create an empty transition system and add some states to it:

    >>> from tulip import transys as trs
    >>> ts = trs.FiniteTransitionSystem()
    >>> ts.states.add('s0')
    >>> ts.states.add_from(['s1', 's3', 'end', 5] )

    Set an initial state, which must already be in states:

    >>> ts.states.initial.add('s0')

    There can be more than one possible initial states:

    >>> ts.states.initial.add_from(['s0', 's3'] )

    To label the states, we need at least one atomic proposition,
    here C{'p'}:

    >>> ts.atomic_propositions |= ['p', None]
    >>> ts.states.add('s0', ap={'p'})
    >>> ts.states.add_from([('s1', {'ap':{'p'} }),
                            ('s3', {'ap':{} } )])

    If a state has already been added, its label of atomic
    propositions can be defined directly:

    >>> ts.states['s0']['ap'] = {'p'}

    Having added states, we can also add some labeled transitions:

    >>> ts.actions |= ['think', 'write']
    >>> ts.transitions.add('s0', 's1', actions='think')
    >>> ts.transitions.add('s1', 5, actions='write')

    Note that an unlabeled transition:

    >>> ts.transitions.add('s0', 's3')

    is considered as different from a labeled one and to avoid
    unintended duplication, after adding an unlabeled transition,
    any attempt to add a labeled transition between the same states
    will raise an exception, unless the unlabeled transition is
    removed before adding the labeled transition.

    The user can still invoke NetworkX functions to set custom node
    and edge labels, in addition to the above ones.
    For example:

    >>> ts.states.add('s0')
    >>> ts.node['s0']['my_cost'] = 5

    The difference is that atomic proposition and action labels
    are checked to make sure they are elements of the system's
    AP and Action sets.

    It is not advisable to use C{MultiDiGraph.add_node} and
    C{MultiDiGraph.add_edge} directly,
    because that can result in an inconsistent system,
    since it skips all checks performed by L{transys}.

    Note
    ====
    The attributes atomic_propositions and aps are equal.
    When you want to produce readable code, use atomic_propositions.
    Otherwise, aps offers shorthand access to the APs.

    Reference
    =========
    For closed systems this corresponds to Def. 2.1, p.20 U{[BK08]
    <http://tulip-control.sourceforge.net/doc/bibliography.html#bk08>}:
        - states (instance of L{States}) = S
        - states.initial = S_0 \subseteq S
        - atomic_propositions = AP
        - actions = Act
        - transitions (instance of L{Transitions})::
              the transition relation ->
                = edge set + edge labeling function
                (labels \in actions)
        Unlabeled edges are defined using:
            - sys.transitions.add
            - sys.transitions.add_from
            - sys.transitions.add_adj
        and accessed using:
            - sys.transitions.find
        - the state labeling function::
                L: S-> 2^AP
        can be defined using:
            - sys.states.add
            - sys.states.add_from
        and accessed using methods:
            - sys.states(data=True)
            - sys.states.find

    See Also
    ========
    L{KripkeStructure}, L{tuple2fts},
    L{line_labeled_with}, L{cycle_labeled_with}
    """

    def __init__(self, env_actions=None, sys_actions=None):
        """Instantiate finite transition system.

        @param env_actions: environment (uncontrolled) actions,
            defined as C{edge_label_types} in L{LabeledDiGraph.__init__}

        @param sys_actions: system (controlled) actions, defined as
            C{edge_label_types} in L{LabeledDiGraph.__init__}
        """
        self._owner = 'sys'

        if env_actions is None:
            env_actions = [
                {'name': 'env_actions',
                 'values': MathSet(),
                 'setter': True}]
        if sys_actions is None:
            sys_actions = [
                {'name': 'sys_actions',
                 'values': MathSet(),
                 'setter': True}]
        # note: "sys_actions" used to be "actions"
        # in closed systems (old FTS)
        action_types = env_actions + sys_actions
        edge_label_types = action_types
        ap_labels = PowerSet()
        node_label_types = [
            {'name': 'ap',
             'values': ap_labels,
             'setter': ap_labels.math_set,
             'default': set()}]
        super(FiniteTransitionSystem, self).__init__(
            node_label_types, edge_label_types)
        # make them available also via an "actions" dicts
        # name, codomain, *rest = x
        actions = {x['name']: x['values'] for x in edge_label_types}
        if 'actions' in actions:
            msg = '"actions" cannot be used as an action type name,\n'
            msg += 'because if an attribute for this action type'
            msg += 'is requested,\n then it will conflict with '
            msg += 'the dict storing all action types.'
            raise ValueError(msg)
        self.actions = actions
        self.atomic_propositions = self.ap
        self.aps = self.atomic_propositions  # shortcut
        # action constraint used in synth.synthesize
        self.env_actions_must = 'xor'
        self.sys_actions_must = 'xor'
        # dot formatting
        self._state_dot_label_format = {
            'ap': '',
            'type?label': '',
            'separator': r'\\n'}
        self._transition_dot_label_format = {
            'sys_actions': 'sys',  # todo: '' if no env
            'env_actions': 'env',
            'type?label': ':',  # todo: '' if no env
            'separator': r'\\n'}
        self._transition_dot_mask = dict()
        self.dot_node_shape = {'normal': 'box'}  # todo: rectangle if no env
        self.default_export_fname = 'fts'

    def __str__(self):
        isopen = (
            ('sys' and any({'env' in x for x in self.actions})) or
            ('env' and any({'sys' in x for x in self.actions})))
        if isopen:
            t = 'open'
        else:
            t = 'closed'
        s = (
            _hl + '\nFinite Transition System (' + t + '): ' +
            self.name + '\n' + _hl + '\n' +
            'Atomic Propositions (APs):\n' +
            pformat(self.atomic_propositions, indent=3) + 2 * '\n' +
            'States labeled with sets of APs:\n' +
            _dumps_states(self) + 2 * '\n' +
            'Initial States:\n' +
            pformat(self.states.initial, indent=3) + 2 * '\n')

        for action_type, codomain in self.actions.iteritems():
            if 'sys' in action_type:
                s += (
                    'System Action Type: ' + str(action_type) +
                    ', with possible values: ' + str(codomain) + '\n' +
                    pformat(codomain, indent=3) + 2 * '\n')
            elif 'env' in action_type:
                s += (
                    'Environment Action Type: ' + str(action_type) +
                    ', with possible values:\n\t' + str(codomain) + '\n' +
                    pformat(codomain, indent=3) + 2 * '\n')
            else:
                s += (
                    'Action type controlled by neither env nor sys\n'
                    ' (will cause you errors later)'
                    ', with possible values:\n\t' +
                    pformat(codomain, indent=3) + 2 * '\n')
        s += (
            'Transitions labeled with sys and env actions:\n' +
            pformat(self.transitions(data=True), indent=3) +
            '\n' + _hl + '\n')
        return s

    @property
    def owner(self):
        return self._owner

    @owner.setter
    def owner(self, x):
        if x not in {'env', 'sys'}:
            raise ValueError("The owner can be either 'sys' or 'env'.")
        self._owner = x

    def _save(self, path, fileformat):
        """Export options available only for closed systems.

        Provides: pml (Promela)

        See Also
        ========
        L{save}, L{plot}
        """
        if fileformat not in {'promela', 'Promela', 'pml'}:
            return False
        # closed ?
        if self.env_vars:
            return False
        from tulip.transys.export import graph2promela
        s = graph2promela.fts2promela(self, self.name)
        # dump to file
        f = open(path, 'w')
        f.write(s)
        f.close()
        return True


class FTS(FiniteTransitionSystem):
    """Alias to L{FiniteTransitionSystem}."""


def tuple2fts(S, S0, AP, L, Act, trans, name='fts',
              prepend_str=None):
    """Create a Finite Transition System from a tuple of fields.

    Hint
    ====
    To remember the arg order:

    1) it starts with states (S0 requires S before it is defined)

    2) continues with the pair (AP, L), because states are more
    fundamental than transitions
    (transitions require states to be defined)
    and because the state labeling L requires AP to be defined.

    3) ends with the pair (Act, trans), because transitions in trans
    require actions in Act to be defined.

    See Also
    ========
    L{tuple2ba}

    @param S: set of states
    @type S: iterable of hashables

    @param S0: set of initial states, must be \\subset S
    @type S0: iterable of elements from S

    @param AP: set of Atomic Propositions for state labeling:
            L: S-> 2^AP
    @type AP: iterable of hashables

    @param L: state labeling definition
    @type L: iterable of (state, AP_label) pairs:
        [(state0, {'p'} ), ...]
        | None, to skip state labeling.

    @param Act: set of Actions for edge labeling:
            R: E-> Act
    @type Act: iterable of hashables

    @param trans: transition relation
    @type trans: list of triples: [(from_state, to_state, act), ...]
        where act \\in Act

    @param name: used for file export
    @type name: str
    """
    def pair_labels_with_states(states, state_labeling):
        if state_labeling is None:
            return
        if not isinstance(state_labeling, Iterable):
            raise TypeError('State labeling function: L->2^AP must be '
                            'defined using an Iterable.')
        state_label_pairs = True
        # cannot be caught by try below
        if isinstance(state_labeling[0], str):
            state_label_pairs = False
        if state_labeling[0] is None:
            state_label_pairs = False
        try:
            (state, ap_label) = state_labeling[0]
        except:
            state_label_pairs = False
        if state_label_pairs:
            return state_labeling
        logger.debug('State labeling L not tuples (state, ap_label),\n'
                     'zipping with states S...\n')
        state_labeling = zip(states, state_labeling)
        return state_labeling
    # args
    if not isinstance(S, Iterable):
        raise TypeError('States S must be iterable, even for single state.')
    # convention
    if not isinstance(S0, Iterable) or isinstance(S0, str):
        S0 = [S0]
    # comprehensive names
    states = S
    initial_states = S0
    ap = AP
    state_labeling = pair_labels_with_states(states, L)
    actions = Act
    transitions = trans
    # prepending states with given str
    if prepend_str:
        logger.debug('Given string:\n\t' + str(prepend_str) + '\n' +
                     'will be prepended to all states.')
    states = prepend_with(states, prepend_str)
    initial_states = prepend_with(initial_states, prepend_str)

    ts = FTS()
    ts.name = name

    ts.states.add_from(states)
    ts.states.initial |= initial_states

    ts.atomic_propositions |= ap

    # note: verbosity before actions below
    # to avoid screening by possible error caused by action

    # state labeling assigned ?
    if state_labeling is not None:
        for state, ap_label in state_labeling:
            if ap_label is None:
                ap_label = set()
            ap_label = str2singleton(ap_label)
            state = prepend_str + str(state)
            logger.debug('Labeling state:\n\t' + str(state) + '\n' +
                         'with label:\n\t' + str(ap_label) + '\n')
            ts.states[state]['ap'] = ap_label
    # any transition labeling ?
    if actions is None:
        for from_state, to_state in transitions:
            (from_state, to_state) = prepend_with([from_state, to_state],
                                                  prepend_str)
            logger.debug('Added unlabeled edge:\n\t' + str(from_state) +
                         '--->' + str(to_state) + '\n')
            ts.transitions.add(from_state, to_state)
    else:
        ts.actions |= actions
        for from_state, to_state, act in transitions:
            (from_state, to_state) = prepend_with([from_state, to_state],
                                                  prepend_str)
            logger.debug(
                'Added labeled edge (=transition):\n\t' +
                str(from_state) + '---[' + str(act) + ']--->' +
                str(to_state) + '\n')
            ts.transitions.add(from_state, to_state, actions=act)
    return ts


def line_labeled_with(L, m=0):
    """Return linear FTS with given labeling.

    The resulting system will be a terminating sequence::
        s0-> s1-> ... -> sN
    where: N = C{len(L) -1}.

    See Also
    ========
    L{cycle_labeled_with}

    @param L: state labeling
    @type L: iterable of state labels, e.g.,::
            [{'p', '!p', 'q',...]
    Single strings are identified with singleton Atomic Propositions,
    so [..., 'p',...] and [...,{'p'},...] are equivalent.

    @param m: starting index
    @type m: int

    @return: L{FTS} with:
        - states ['s0', ..., 'sN'], where N = len(L) -1
        - state labels defined by L, so s0 is labeled with L[0], etc.
        - transitions forming a sequence:
            - s_{i} ---> s_{i+1}, for: 0 <= i < N
    """
    n = len(L)
    S = range(m, m + n)
    S0 = []  # user will define them
    AP = {True}
    for ap_subset in L:
        # skip empty label ?
        if ap_subset is None:
            continue
        AP |= set(ap_subset)
    Act = None
    from_states = range(m, m + n - 1)
    to_states = range(m + 1, m + n)
    trans = zip(from_states, to_states)
    ts = tuple2fts(S, S0, AP, L, Act, trans, prepend_str='s')
    return ts


def cycle_labeled_with(L):
    """Return cycle FTS with given labeling.

    The resulting system will be a cycle::
        s0-> s1-> ... -> sN -> s0
    where: N = C{len(L) -1}.

    See Also
    ========
    L{line_labeled_with}

    @param L: state labeling
    @type L: iterable of state labels, e.g., [{'p', 'q'}, ...]
        Single strings are identified with singleton Atomic Propositions,
        so [..., 'p',...] and [...,{'p'},...] are equivalent.

    @return: L{FTS} with:
        - states ['s0', ..., 'sN'], where N = len(L) -1
        - state labels defined by L, so s0 is labeled with L[0], etc.
        - transitions forming a cycle:
            - s_{i} ---> s_{i+1}, for: 0 <= i < N
            - s_N ---> s_0
    """
    ts = line_labeled_with(L)
    last_state = 's' + str(len(L) - 1)
    ts.transitions.add(last_state, 's0')
    # trans += [(n-1, 0)] # close cycle
    return ts


def add_initial_states(ts, ap_labels):
    """Make initial any state of ts labeled with any label in ap_labels.

    For example if isinstance(ofts, FTS):

      >>> from tulip.transys.transys import add_initial_states
      >>> initial_labels = [{'home'}]
      >>> add_initial_states(ofts, initial_labels)

    @type ts: L{FiniteTransitionSystem}

    @param ap_labels: labels, each comprised of atomic propositions
    @type ap_labels: iterable of sets of elements from
        ts.atomic_propositions
    """
    for label in ap_labels:
        new_init_states = ts.states.find(ap='label')
        ts.states.initial |= new_init_states


def _dumps_states(g):
    """Dump string of transition system states.

    @type g: L{FTS}
    """
    nodes = g
    a = []
    for u in nodes:
        s = '\t State: {u}, AP: {ap}\n'.format(
            u=u, ap=g.node[u]['ap']) + ', '.join([
                '{k}: {v}'.format(k=k, v=v)
                for k, v in g.node[u].iteritems()
                if k is not 'ap'])
        a.append(s)
    return ''.join(a)


class GameGraph(LabeledDiGraph):
    """Store a game graph.

    When adding states, you have to say
    which player controls the outgoing transitions.
    Use C{networkx} state labels for that:

      >>> g = GameGraph()
      >>> g.states.add('s0', player=0)

    See also
    ========
    L{automata.ParityGame}

    Reference
    =========
    1. Chatterjee K.; Henzinger T.A.; Jobstmann B.
       Environment Assumptions for Synthesis
       CONCUR'08, LNCS 5201, pp. 147-161, 2008
    """

    def __init__(self, node_label_types, edge_label_types):
        node_label_types += [{
            'name': 'player',
            'values': {0, 1},
            'default': 0}]
        super(GameGraph, self).__init__(node_label_types,
                                        edge_label_types)

    def player_states(self, n):
        """Return states controlled by player C{n}.

        'controlled' means that player C{n}
        gets to decide the successor state.

        @param n: player index (id number)
        @type n: 0 or 1

        @return: set of states
        @rtype: C{set}
        """
        return {x for x in self if self.node[x]['player'] == n}

    def edge_controlled_by(self, e):
        """Return the index of the player controlling edge C{e}.

        @type e: 2-tuple of nodes C{(n1, n2)}

        @rtype: integer 0 or 1
        """
        from_state = e[0]
        return self.node[from_state]['player']


class LabeledGameGraph(GameGraph):
    """Game graph with labeled states.

    Its contraction is a Kripke structure.
    Given a Kripke structure and a partition of propositions,
    then the corresponding labeled game graph
    can be obtained by graph expansion.

    Reference
    =========
    1. Chatterjee K.; Henzinger T.A.; Piterman N.
       Strategy Logic
       UCB/EECS-2007-78
    """

    def __init__(self):
        ap_labels = PowerSet()
        node_label_types = [
            {'name': 'ap',
             'values': ap_labels,
             'setter': ap_labels.math_set,
             'default': set()}]
        super(LabeledGameGraph, self).__init__(node_label_types)
        self.atomic_propositions = self.ap
        # dot formatting
        self._state_dot_label_format = {
            'ap': '',
            'type?label': '',
            'separator': r'\\n'}
        self.dot_node_shape = {'normal': 'rectangle'}



# Copyright (c) 2013 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""
Export html containing d3.js animating SVG
"""
# there exists: https://github.com/mikedewar/d3py,
# but it is not sufficiently developed yet,
# so here the wheel is partially re-invented

import os
import inspect

from networkx.readwrite import json_graph

def _format_label(label_def, label_dot_format):
    """Format state/edge labels, which pop-up on mouse hover.
    """
    s = '"\\n\\n" +'
    for sublabel_name in label_def:
        shown_name = label_dot_format[sublabel_name]
        kv_sep = label_dot_format['type?label']
        sep = label_dot_format['separator']

        s += '"' +shown_name +kv_sep +'" '
        s += '+d.' +str(sublabel_name) +'+"' +sep+'" +'
    s += '" ";'

    return s

def labeled_digraph2d3(graph, html_file_name='index.html'):
    """Export to SVG embedded in HTML, animated with d3.js

    Example
    =======
    From C{examples/transys/machine_examples.py} call:

    >>> m = garage_counter_with_state_vars()

    Then export to html:

    >>> m.save('index.html', 'html')

    See Also
    ========
    FSM, BA, Mealy

    @param graph: labeled graph to export
    @type graph: L{LabeledDiGraph}
    """
    file_path = inspect.getfile(inspect.currentframe())
    dir_path = os.path.dirname(os.path.abspath(file_path) )

    d3_file_name = os.path.join(dir_path, 'd3.v3.min.js')
    d3_file = open(d3_file_name)
    d3_js = d3_file.read()

    s = """
    <!DOCTYPE html>
    <meta charset="utf-8">
    <style>

    .node {
      stroke: black;
      stroke-width: 1.5px;
    }

    .link {
      stroke: #999;
      stroke-opacity: .6;
    }

    .end-arrow {
        fill            : gray;
        stroke-width    : 1px;
    }

    </style>

    <script>
    """

    # embed d3.js to create single .html,
    # instead of bunch of files
    s += d3_js

    s += """
    </script>
    <body>

    <script>
    var width = 960,
        height = 500;

    var color = d3.scale.category20();

    var force = d3.layout.force()
        .charge(-120)
        .linkDistance(200)
        .size([width, height]);

    var svg = d3.select("body").append("svg")
        .attr("width", width)
        .attr("height", height);

    svg.append('svg:defs').append('svg:marker')
        .attr('id', 'end-arrow')
        .attr('viewBox', '0 -5 10 10')
        .attr('refX', 15)
        .attr('refY', 0)
        .attr('markerWidth', 4)
        .attr('markerHeight', 4)
        .attr('orient', 'auto')
      .append('svg:path')
        .attr('d', 'M0,-5L10,0L0,5')
        .attr('class', 'end-arrow');

    var graph = """

    # embed to avoid browser local file-loading restrictions
    try:
        s += json_graph.dumps(graph)
    except:
        # better error msg for numpy array
        import json
        data = json_graph.node_link_data(graph)
        s += json.dumps(data, default=lambda x: str(x) )

    s += ';'

    s += """
    function draw(graph){
      force
          .nodes(graph.nodes)
          .links(graph.links)
          .start();

      var link = svg.append("svg:g").selectAll("path")
          .data(graph.links)
        .enter().append("svg:path")
          .attr("class", "link")
          .style("stroke-width", 10)
          .style("fill", "none")
          .style("marker-end", 'url(#end-arrow)');

      link.append("title")
          .text(function(d) {
          	return """

    # edge labels (shown when mouse on edge)
    if hasattr(graph, '_transition_label_def') and \
    hasattr(graph, '_transition_dot_label_format'):
        transition_label_def = graph._transition_label_def
        transition_label_format = graph._transition_dot_label_format
        s += _format_label(transition_label_def, transition_label_format)
    else:
        s += '" ";'

    s += """});

      var node = svg.selectAll(".node")
          .data(graph.nodes)
       .enter().append("g")
          .attr("class", "node")
          .call(force.drag);

      node.append("circle")
          .attr("r", 30)
          .style("fill", "#66CC00")

      node.append("text")
          .attr("dx", 0)
          .attr("dy", 0)
          .attr("fill", "red")
          .text(function(d) { return d.id});

      node.append("title")
          .style("fill", "gray")
          .text(function(d) { return """

    # edge labels (shown when mouse on edge)
    if hasattr(graph, '_state_label_def') and \
    hasattr(graph, '_state_dot_label_format'):
        state_label_def = graph._state_label_def
        state_label_format = graph._state_dot_label_format
        s += _format_label(state_label_def, state_label_format)
    else:
        s += '" ";'

    s += """});

      force.on("tick", function() {
        link.attr("d", function(d) {
            var dx = d.target.x -d.source.x,
                dy = d.target.y -d.source.y,
                dr = Math.sqrt(dx * dx + dy * dy);
            return "M" +
                d.source.x + "," +
                d.source.y + "A" +
                dr + "," + dr + " 0 0,1 " +
                d.target.x + "," +
                d.target.y;
        });

        node.attr("cx", function(d) { return d.x; })
            .attr("cy", function(d) { return d.y; })
            .attr("transform", function(d) {
                return "translate(" + d.x + "," + d.y + ")";
            });
      });
    };

    draw(graph)

    </script>
    </body>
    """

    html_file = open(html_file_name, 'w')
    html_file.write(s)
    return True



#!/usr/bin/env python

"""
This example shows how to load from a json file
a configuration for logging messages from tulip.

You can configure it also directly in python, but
rewriting or copy/pasting again and again the same
base configuration is not very efficient.
"""
import logging.config
import json
import pprint

fname = 'logging_config.json'
with open(fname, 'r') as f:
    config = json.load(f)
print('Your logging config is:\n{s}')
pprint.pprint(config)

logging.config.dictConfig(config)

from tulip import transys, spec, synth

sys = transys.FTS()
sys.states.add_from({0, 1})
sys.states.initial.add(0)

sys.add_edges_from([(0, 1), (1, 0)])

sys.atomic_propositions.add('p')
sys.node[0]['ap'] = {'p'}

specs = spec.GRSpec(sys_vars={'p'}, sys_prog={'p'})

mealy = synth.synthesize('gr1c', specs, sys=sys)



"""
Tests for the tulip.synth module.
"""
import logging
logging.getLogger('tulip').setLevel(logging.ERROR)
logging.getLogger('tulip.interfaces.omega').setLevel(logging.DEBUG)
logging.getLogger('omega').setLevel(logging.WARNING)
from nose.tools import assert_raises
import numpy as np
from scipy import sparse as sp
from tulip import spec, synth, transys


def sys_fts_2_states():
    sys = transys.FTS()
    sys.states.add_from(['X0', 'X1'])
    sys.states.initial.add_from(['X0', 'X1'])

    sys.transitions.add('X0', 'X1')
    sys.transitions.add('X1', 'X0')
    sys.transitions.add('X1', 'X1')

    sys.atomic_propositions.add_from({'home', 'lot'})
    sys.states.add('X0', ap={'home'})
    sys.states.add('X1', ap={'lot'})

    # sys.plot()
    return sys


def env_fts_2_states():
    env = transys.FTS()
    env.owner = 'env'

    env.states.add_from({'e0', 'e1'})
    env.states.initial.add('e0')

    # Park as an env action
    env.env_actions.add_from({'park', 'go'})

    env.transitions.add('e0', 'e0', env_actions='park')
    env.transitions.add('e0', 'e0', env_actions='go')

    # env.plot()
    return env


def env_ofts_bool_actions():
    env = transys.FTS()
    env.owner = 'env'

    env.states.add_from({'e0', 'e1', 'e2'})
    env.states.initial.add('e0')

    env.env_actions.add_from({'park', 'go'})
    env.sys_actions.add_from({'up', 'down'})

    env.transitions.add('e0', 'e1', env_actions='park',
                        sys_actions='up')
    env.transitions.add('e1', 'e2', env_actions='go',
                        sys_actions='down')
    # env.plot()
    return env


def env_ofts_int_actions():
    env = env_ofts_bool_actions()

    env.env_actions.add('stop')
    env.sys_actions.add('hover')

    return env


def parking_spec():
    # barely realizable: assumption necessary
    env_prog = '! (eact = park)'

    sys_vars = {'X0reach'}
    sys_init = {'X0reach'}
    sys_prog = {'home'}

    # one additional requirement: if in lot,
    # then stay there until park signal is turned off
    sys_safe = {'(X (X0reach) <-> lot) || (X0reach && !(eact = "park") )',
                '((lot & (eact = "park") ) -> X(lot))'}
    sys_prog |= {'X0reach'}

    specs = spec.GRSpec(sys_vars=sys_vars, sys_init=sys_init,
                        sys_safety=sys_safe,
                        env_prog=env_prog, sys_prog=sys_prog)
    return specs


def test_sys_fts_int_states():
    """Sys FTS has 3 states, must become 1 int var in GR(1).
    """
    sys = sys_fts_2_states()
    sys.sys_actions_must = 'mutex'
    sys.states.add('X2')

    spec = synth.sys_to_spec(
        sys,
        ignore_initial=False,
        statevar='loc',
        bool_actions=False
    )

    assert 'X0' not in spec.sys_vars
    assert 'X1' not in spec.sys_vars
    assert 'X2' not in spec.sys_vars

    assert 'eloc' not in spec.sys_vars
    assert 'loc' in spec.sys_vars
    assert sorted(spec.sys_vars['loc']) == ['X0', 'X1', 'X2']


def test_env_fts_int_states():
    """Env FTS has 3 states, must become 1 int var in GR(1).
    """
    env = env_fts_2_states()
    env.env_actions_must = 'mutex'
    env.states.add('e2')

    spec = synth.env_to_spec(
        env,
        ignore_initial=False,
        statevar='eloc',
        bool_actions=False
    )

    assert 'e0' not in spec.env_vars
    assert 'e1' not in spec.env_vars
    assert 'e2' not in spec.env_vars

    assert 'loc' not in spec.env_vars
    assert 'eloc' in spec.env_vars
    print(spec.env_vars['eloc'])
    assert sorted(spec.env_vars['eloc']) == ['e0', 'e1', 'e2']


def test_sys_fts_no_actions():
    """Sys FTS has no actions."""
    sys = sys_fts_2_states()
    sys.sys_actions_must = 'mutex'

    spec = synth.sys_to_spec(
        sys,
        ignore_initial=False,
        statevar='loc',
        bool_actions=False)
    assert 'actions' not in spec.sys_vars


def test_env_fts_bool_actions():
    """Env FTS has 2 actions, bools requested."""
    env = env_fts_2_states()
    env.env_actions_must = 'mutex'

    spec = synth.env_to_spec(
        env,
        ignore_initial=False,
        statevar='eloc',
        bool_actions=True,
    )

    assert 'sys_actions' not in spec.env_vars
    assert 'env_actions' not in spec.env_vars

    assert 'park' in spec.env_vars
    assert spec.env_vars['park'] == 'boolean'

    assert 'go' in spec.env_vars
    assert spec.env_vars['go'] == 'boolean'


def test_env_fts_int_actions():
    """Env FTS actions must become 1 int var in GR(1).
    """
    env = env_fts_2_states()
    env.env_actions_must = 'mutex'
    env.env_actions.add('stop')

    spec = synth.env_to_spec(
        env,
        ignore_initial=False,
        statevar='eloc',
        bool_actions=False
    )

    assert 'park' not in spec.env_vars
    assert 'go' not in spec.env_vars
    assert 'stop' not in spec.env_vars

    assert 'sys_actions' not in spec.env_vars
    assert 'env_actions' in spec.env_vars

    print spec.env_vars['env_actions']
    assert (set(spec.env_vars['env_actions']) ==
            {'park', 'go', 'stop', 'env_actionsnone'})


def test_env_ofts_bool_actions():
    """Env OpenFTS has 2 actions, bools requested.
    """
    env = env_ofts_int_actions()
    env.env_actions_must = 'mutex'
    env.env_actions.remove('stop')
    env.sys_actions.remove('hover')

    spec = synth.env_to_spec(
        env,
        ignore_initial=False,
        statevar='eloc',
        bool_actions=True
    )

    _check_ofts_bool_actions(spec)


def test_sys_ofts_bool_actions():
    """Sys OpenFTS has 2 actions, bools requested.
    """
    sys = env_ofts_int_actions()
    sys.owner = 'sys'
    sys.sys_actions_must = 'mutex'
    sys.env_actions.remove('stop')
    sys.sys_actions.remove('hover')

    spec = synth.sys_to_spec(
        sys,
        ignore_initial=False,
        statevar='loc',
        bool_actions=True
    )

    _check_ofts_bool_actions(spec)


def _check_ofts_bool_actions(spec):
    """Common assertion checking for 2 functions above."""
    assert 'park' in spec.env_vars
    assert spec.env_vars['park'] == 'boolean'

    assert 'go' in spec.env_vars
    assert spec.env_vars['go'] == 'boolean'

    assert 'up' in spec.sys_vars
    assert spec.sys_vars['up'] == 'boolean'

    assert 'down' in spec.sys_vars
    assert spec.sys_vars['down'] == 'boolean'

    assert 'env_actions' not in spec.env_vars
    assert 'sys_actions' not in spec.sys_vars


def test_env_ofts_int_actions():
    """Env OpenFTS actions must become 1 int var in GR(1)."""
    env = env_ofts_int_actions()
    env.sys_actions_must = 'mutex'
    spec = synth.env_to_spec(
        env,
        ignore_initial=False,
        statevar='eloc',
        bool_actions=False)
    _check_ofts_int_actions(spec)


def test_sys_ofts_int_actions():
    """Sys OpenFTS actions must become 1 int var in GR(1).
    """
    sys = env_ofts_int_actions()
    sys.owner = 'sys'
    sys.sys_actions_must = 'mutex'
    spec = synth.sys_to_spec(
        sys,
        ignore_initial=False,
        statevar='loc',
        bool_actions=False)
    _check_ofts_int_actions(spec)


def _check_ofts_int_actions(spec):
    """Common assertion checking for 2 function above."""
    print(spec.env_vars)
    print(spec.sys_vars)
    assert 'park' not in spec.env_vars
    assert 'go' not in spec.env_vars
    assert 'stop' not in spec.env_vars

    assert 'up' not in spec.sys_vars
    assert 'down' not in spec.sys_vars
    assert 'hover' not in spec.sys_vars

    assert 'env_actions' in spec.env_vars
    assert set(spec.env_vars['env_actions']) == {'park', 'go', 'stop'}

    assert 'sys_actions' in spec.sys_vars
    assert (set(spec.sys_vars['sys_actions']) == {'up', 'down', 'hover',
            'sys_actionsnone'})


def test_only_mode_control():
    """Unrealizable due to non-determinism.

    Switched system with 2 modes: 'left', 'right'.
    Modes are controlled by the system.
    States are controlled by the environment.

    So only control over dynamics is through mode switching.
    Transitions are thus interpreted as non-deterministic.

    This can model uncertain outcomes in the real world,
    e.g., due to low quality actuators or
    bad low-level feedback controllers.
    """
    # Create a finite transition system
    env_sws = transys.FTS()
    env_sws.owner = 'env'

    env_sws.sys_actions.add_from({'right', 'left'})

    # str states
    n = 4
    states = transys.prepend_with(range(n), 's')

    env_sws.atomic_propositions.add_from(['home', 'lot'])

    # label TS with APs
    ap_labels = [set(), set(), {'home'}, {'lot'}]
    for i, label in enumerate(ap_labels):
        state = 's' + str(i)
        env_sws.states.add(state, ap=label)

    # mode1 transitions
    transmat1 = np.array([[0, 1, 0, 1],
                          [0, 1, 0, 0],
                          [0, 1, 0, 1],
                          [0, 0, 0, 1]])
    env_sws.transitions.add_adj(
        sp.lil_matrix(transmat1), states, {'sys_actions': 'right'})

    # mode2 transitions
    transmat2 = np.array([[1, 0, 0, 0],
                          [1, 0, 1, 0],
                          [0, 0, 1, 0],
                          [1, 0, 1, 0]])
    env_sws.transitions.add_adj(
        sp.lil_matrix(transmat2), states, {'sys_actions': 'left'})

    env_vars = {'park'}
    env_init = {'eloc = "s0"', 'park'}
    env_prog = {'!park'}
    env_safe = set()

    sys_vars = {'X0reach'}
    sys_init = {'X0reach'}
    sys_prog = {'home'}
    sys_safe = {'next(X0reach) <-> lot || (X0reach && !park)'}
    sys_prog |= {'X0reach'}

    specs = spec.GRSpec(env_vars, sys_vars, env_init, sys_init,
                        env_safe, sys_safe, env_prog, sys_prog)

    r = synth.is_realizable('omega', specs, env=env_sws, ignore_env_init=True)
    assert not r


def multiple_env_actions_test():
    multiple_env_actions_check('omega')

def multiple_env_actions_check(solver='omega'):
    """Two env players, 3 states controlled by sys.

    sys wins marginally, due to assumption on
    next combination of actions by env players.
    """
    # 1 <---> 2
    # 1  ---> 3

    env_actions = [
        {
            'name': 'env_alice',
            'values': transys.MathSet({'left', 'right'})
        },
        {
            'name': 'env_bob',
            'values': transys.MathSet({'bleft', 'bright'})
        }
    ]

    sys = transys.FTS(env_actions)
    sys.states.add_from({'s1', 's2', 's3'})
    sys.states.initial.add_from({'s1'})

    sys.add_edge('s1', 's2', env_alice='left', env_bob='bright')
    sys.add_edge('s2', 's1', env_alice='left', env_bob='bright')
    # at state 3 sys loses
    sys.add_edge('s1', 's3', env_alice='right', env_bob='bleft')

    logging.debug(sys)

    env_safe = {('(loc = "s1") -> X( (env_alice = "left") && '
                 '(env_bob = "bright") )')}
    sys_prog = {'loc = "s1"', 'loc = "s2"'}

    specs = spec.GRSpec(
        env_safety=env_safe,
        sys_prog=sys_prog,
        moore=False,
        plus_one=False,
        qinit='\A \E')
    r = synth.is_realizable(solver, specs, sys=sys)
    assert r
    # slightly relax assumption
    specs = spec.GRSpec(
        sys_prog=sys_prog,
        moore=False,
        plus_one=False,
        qinit='\A \E')
    r = synth.is_realizable(solver, specs, sys=sys)
    assert not r


def test_var_name_conflicts():
    """Check redefinitions between states, actions, atomic props."""
    conversion_raises = lambda x, y: assert_raises(
        Exception, spec=x, sys=y,
        ignore_initial=True,
        statevar='loc',
        bool_actions=False)

    # FTS to spec

    # states vs APs
    sys = transys.FTS()
    sys.states.add('out')
    sys.atomic_propositions.add('out')

    conversion_raises(synth.sys_to_spec, sys)

    # states vs sys_actions
    sys = transys.FTS()
    sys.states.add('out')
    sys.sys_actions.add('out')

    conversion_raises(synth.sys_to_spec, sys)

    sys = transys.FTS()
    sys.states.add('sys_actions')
    sys.sys_actions.add('out')

    conversion_raises(synth.sys_to_spec, sys)

    # states vs env_actions
    env = transys.FTS()
    env.states.add('out')
    env.env_actions.add('out')

    conversion_raises(synth.env_to_spec, env)

    env = transys.FTS()
    env.states.add('env_actions')
    env.env_actions.add('out')

    conversion_raises(synth.env_to_spec, env)

    # APs vs sys_actions
    sys = transys.FTS()
    sys.states.add('s0')
    sys.atomic_propositions.add('out')
    sys.env_actions.add('out')

    conversion_raises(synth.sys_to_spec, sys)

    sys = transys.FTS()
    sys.states.add('s0')
    sys.atomic_propositions.add('sys_actions')
    sys.env_actions.add('out')

    conversion_raises(synth.sys_to_spec, sys)

    # APs vs env_actions
    env = transys.FTS()
    env.states.add('s0')
    env.atomic_propositions.add('out')
    env.env_actions.add('out')

    conversion_raises(synth.env_to_spec, env)

    env = transys.FTS()
    env.states.add('s0')
    env.atomic_propositions.add('env_actions')
    env.env_actions.add('out')

    conversion_raises(synth.env_to_spec, env)

    # OpenFTS to spec

    # states vs APs
    sys = transys.FTS()
    sys.states.add('out')
    sys.atomic_propositions.add('out')

    conversion_raises(synth.sys_to_spec, sys)

    conversion_raises(synth.env_to_spec, sys)

    # states vs sys_actions
    sys = transys.FTS()
    sys.states.add('out')
    sys.sys_actions.add('out')

    conversion_raises(synth.sys_to_spec, sys)

    conversion_raises(synth.env_to_spec, sys)

    sys = transys.FTS()
    sys.states.add('sys_actions')
    sys.sys_actions.add('out')

    conversion_raises(synth.sys_to_spec, sys)

    conversion_raises(synth.env_to_spec, sys)

    # states vs env_actions
    sys = transys.FTS()
    sys.states.add('out')
    sys.env_actions.add('out')

    conversion_raises(synth.sys_to_spec, sys)

    conversion_raises(synth.env_to_spec, sys)

    sys = transys.FTS()
    sys.states.add('env_actions')
    sys.env_actions.add('out')

    conversion_raises(synth.sys_to_spec, sys)

    conversion_raises(synth.env_to_spec, sys)

    # sys_actions vs APs
    sys = transys.FTS()
    sys.states.add('s0')
    sys.sys_actions.add('out')
    sys.atomic_propositions.add('out')

    conversion_raises(synth.sys_to_spec, sys)

    conversion_raises(synth.env_to_spec, sys)

    sys = transys.FTS()
    sys.states.add('s0')
    sys.sys_actions.add('out')
    sys.atomic_propositions.add('sys_actions')

    conversion_raises(synth.sys_to_spec, sys)

    conversion_raises(synth.env_to_spec, sys)

    # env_actions vs APs
    sys = transys.FTS()
    sys.states.add('s0')
    sys.env_actions.add('out')
    sys.atomic_propositions.add('out')

    conversion_raises(synth.sys_to_spec, sys)

    conversion_raises(synth.env_to_spec, sys)

    sys = transys.FTS()
    sys.states.add('s0')
    sys.env_actions.add('out')
    sys.atomic_propositions.add('env_actions')

    conversion_raises(synth.sys_to_spec, sys)

    conversion_raises(synth.env_to_spec, sys)


def test_determinize_machine_init():
    mach = transys.MealyMachine()
    mach.add_inputs({'a': {0, 1}})
    mach.add_outputs({'b': {0, 1}, 'c': {0, 1}})
    u = 'Sinit'
    mach.add_nodes_from([u, 1, 2])

    # initial reactions:

    # to input: a=0
    mach.add_edge(u, 1, a=0, b=0, c=1)
    mach.add_edge(u, 2, a=0, b=0, c=1)
    mach.add_edge(u, 1, a=0, b=1, c=0)
    mach.add_edge(u, 2, a=0, b=1, c=1)

    # to input: a=1
    mach.add_edge(u, 1, a=1, b=0, c=1)
    mach.add_edge(u, 2, a=1, b=0, c=1)
    mach.add_edge(u, 1, a=1, b=1, c=0)
    mach.add_edge(u, 2, a=1, b=1, c=1)

    # determinize all outputs arbitrarily
    detmach = synth.determinize_machine_init(mach)
    assert detmach is not mach

    for a in {0, 1}:
        edges = [(i, j) for (i, j, d) in detmach.edges_iter(u, data=True)
                 if d['a'] == a]
        assert len(edges) == 1

    # determinize output b arbitrarily,
    # but output c is constrained to the initial value 0
    detmach = synth.determinize_machine_init(mach, {'c': 0})

    for a in {0, 1}:
        edges = [(i, j, d) for (i, j, d) in detmach.edges_iter(u, data=True)
                 if d['a'] == a]
        assert len(edges) == 1

        ((i, j, d), ) = edges
        assert j == 1
        assert d['b'] == 1


class synthesize_test:
    def setUp(self):
        self.f_triv = spec.GRSpec(
            sys_vars="y",
            moore=False,
            plus_one=False)
        self.trivial_unreachable = spec.GRSpec(
            sys_vars="y",
            sys_prog="False",
            moore=False,
            plus_one=False)

    def tearDown(self):
        self.f_triv = None

    def test_gr1c_basic(self):
        g = synth.synthesize("omega", self.f_triv)
        assert isinstance(g, transys.MealyMachine)

    def test_unrealizable(self):
        assert synth.synthesize("omega", self.trivial_unreachable) is None


if __name__ == '__main__':
    multiple_env_actions_test()



"""Usage example for the function `abstract.get_input`.

To be run after `continuous.py`, in same session.
For example, within an `ipython` interactive session in
this directory:

run ../continuous.py
run -i test_get_input.py
"""
from tulip.abstract import get_input, find_discrete_state
from polytope import is_inside
import numpy as np

def integrate(sys_dyn, x0, u_seq):
    # is the continuous transition correct ?
    N = u_seq.shape[0]
    x = x0.reshape(x0.size, 1)

    A = sys_dyn.A
    B = sys_dyn.B

    if len(sys_dyn.K) == 0:
        K = np.zeros(x.shape)
    else:
        K = sys_dyn.K

    print('started continuous transition')
    m = u_seq[0, :].size
    for i in xrange(N):
        u = u_seq[i, :].reshape(m, 1)
        x = A.dot(x) + B.dot(u) + K

        print('Discrete time: k = ' +str(i) )
        print('\t u[' +str(i) +"]' = " +str(u.T) )
        print('\t x[' +str(i) +"]' = " +str(x.T) +'\n')

    print('completed continuous transition iteration')
    return x

x0 = np.array([0.5, 0.6])
start = find_discrete_state(x0, disc_dynamics.ppp)
end = 14

start_poly = disc_dynamics.ppp.regions[start]
end_poly = disc_dynamics.ppp.regions[end]

if not is_inside(start_poly, x0):
    raise Exception('x0 \\notin start_poly')

start_state = start
end_state = end

post = disc_dynamics.ts.states.post(start_state)
print(post)
if not end_state in post:
    raise Exception('end \\notin post(start)')

u_seq = get_input(x0, sys_dyn, disc_dynamics,
              start, end)
print('Computed input sequence: u = ')
print(u_seq)

x = integrate(sys_dyn, x0, u_seq)

# arrived at target ?
if not is_inside(end_poly, x):
    raise Exception('incorrect continuous transition')
else:
    print('arrived at target Region/Polytope')



"""Tests for interface to `omega.games.gr1`."""
import logging

import networkx as nx

from tulip.spec import form
from tulip.interfaces import omega as omega_int
from tulip import synth


from nose import tools as nt


logging.getLogger('tulip').setLevel('ERROR')
logging.getLogger('astutils').setLevel('ERROR')
logging.getLogger('omega').setLevel('ERROR')
log = logging.getLogger('omega.games')
log.setLevel('WARNING')
log.addHandler(logging.StreamHandler())


def test_grspec_to_automaton():
    sp = grspec_0()
    a = omega_int._grspec_to_automaton(sp)
    dvars = dict(x=dict(type='bool', owner='env', dom=None),
                 y=dict(type='bool', owner='sys', dom=None))
    assert a.vars == dvars, (a.vars, dvars)
    r = a.action['env']
    assert r == list(), r
    r = a.action['sys']
    assert r == ['( ( X x ) -> ( X y ) )'], r
    r = a.win['<>[]']
    assert r == ['!(( ! x ))'], r
    r = a.win['[]<>']
    assert r == ['( ! y )'], r


def test_synthesis_bool():
    sp = grspec_0()
    h = omega_int.synthesize_enumerated_streett(sp)
    assert h is not None, 'no winning states'
    g = synth.strategy2mealy(h, sp)
    # fname = 'mealy.pdf'
    # g.save(fname)
    # pd = nx.drawing.nx_pydot.to_pydot(g)
    # pd.write_pdf(fname)
    assert g is not None
    assert len(g.inputs) == 1, g.inputs
    assert 'x' in g.inputs, g.inputs
    dom = g.inputs['x']
    dom_ = {False, True}
    assert dom == dom_, (dom, dom_)
    assert len(g.outputs) == 1, g.outputs
    assert 'y' in g.outputs, g.outputs
    dom = g.outputs['y']
    dom_ = {False, True}
    assert dom == dom_, (dom, dom_)
    assert len(g) == 5, [
        g.nodes(data=True), g.edges(data=True)]


def test_synthesis_fol():
    sp = grspec_1()
    h = omega_int.synthesize_enumerated_streett(sp)
    assert h is not None
    g = synth.strategy2mealy(h, sp)
    assert g is not None
    assert len(g.inputs) == 1, g.inputs
    assert 'x' in g.inputs, g.inputs
    dom = g.inputs['x']
    dom_ = set(xrange(5))
    assert dom == dom_, (dom, dom_)
    assert len(g.outputs) == 1, g.outputs
    assert 'y' in g.outputs, g.outputs
    dom = g.outputs['y']
    dom_ = set(xrange(5))
    assert dom == dom_, (dom, dom_)


def test_synthesis_strings():
    sp = grspec_4()
    h = omega_int.synthesize_enumerated_streett(sp)
    g = synth.strategy2mealy(h, sp)
    assert g is not None
    # outputs
    assert len(g.outputs) == 1, g.outputs
    assert 'y' in g.outputs, g.outputs
    dom = g.outputs['y']
    dom_ = {'a', 'b'}
    assert dom == dom_, (dom, dom_)
    # check simulation
    n = len(g)
    assert n == 4, n
    u = 'Sinit'
    u, r = g.reaction(u, dict(x=0))
    assert r == dict(y='a'), r
    u, r = g.reaction(u, dict(x=2))
    assert r == dict(y='b'), r
    u, r = g.reaction(u, dict(x=1))
    assert r == dict(y='b'), r
    u, r = g.reaction(u, dict(x=0))
    assert r == dict(y='a'), r


def test_synthesis_moore():
    sp = grspec_2()
    h = omega_int.synthesize_enumerated_streett(sp)
    g = synth.strategy2mealy(h, sp)
    # g.save('moore.pdf')
    assert g is not None
    n = len(g)
    assert n == 26, n


def test_synthesis_mealy_all_init():
    sp = grspec_3()
    h = omega_int.synthesize_enumerated_streett(sp)
    assert h is None, 'should be unrealizable'
    sp.env_init = ['y = x']
    h = omega_int.synthesize_enumerated_streett(sp)
    g = synth.strategy2mealy(h, sp)
    # g.save('moore.pdf')
    assert g is not None
    n = len(g)
    assert n == 6, n


def test_synthesis_unrealizable():
    sp = grspec_0()
    sp.sys_prog = ['False']
    h = omega_int.synthesize_enumerated_streett(sp)
    assert h is None, h


def test_is_circular_true():
    f = form.GRSpec()
    f.sys_vars['y'] = 'bool'
    f.env_prog = ['y']
    f.sys_prog = ['y']
    triv = omega_int.is_circular(f)
    assert triv, triv


def test_is_circular_false():
    f = form.GRSpec()
    f.env_vars['x'] = 'bool'
    f.env_prog = ['x']
    f.sys_prog = ['x']
    triv = omega_int.is_circular(f)
    assert not triv, triv


def grspec_0():
    sp = form.GRSpec()
    sp.moore = False
    sp.env_vars = dict(x='boolean')
    sp.sys_vars = dict(y='boolean')
    sp.sys_safety = ["x' -> y'"]
    sp.env_prog = ['!x']
    sp.sys_prog = ['!y']
    return sp


def grspec_1():
    sp = form.GRSpec()
    sp.moore = False
    sp.plus_one = False
    sp.env_vars = dict(x=(0, 4))
    sp.sys_vars = dict(y=(0, 4))
    sp.env_init = ['(0 <= y) & (y <= 4)']
    sp.sys_safety = ["y' = x'"]
    return sp


def grspec_2():
    sp = form.GRSpec()
    sp.moore = True
    sp.env_vars = dict(x=(0, 4))
    sp.sys_vars = dict(y=(0, 4))
    sp.env_init = ['(0 <= y) & (y <= 4)']
    sp.sys_safety = ["y' = x"]
    return sp

def grspec_3():
    sp = form.GRSpec()
    sp.moore = False
    sp.env_vars = dict(x=(0, 4))
    sp.sys_vars = dict(y=(0, 4))
    sp.env_init = ['(0 <= y) & (y <= 4)']
    sp.sys_safety = ["y = x"]
    return sp


def grspec_4():
    sp = form.GRSpec()
    sp.moore = False
    sp.env_init = ['(x = 0) & (y = "a")']
    sp.env_vars = dict(x=(0, 2))
    sp.sys_vars = dict(y=['a', 'b'])
    sp.sys_safety = [
        '(x = 0) -> (y = "a")',
        '(x > 0) -> (y = "b")']
    return sp


if __name__ == '__main__':
    test_synthesis_bool()



# Copyright (c) 2012-2016 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Routines for working with gridworlds.

Note (24 June 2012): Several pieces of source code are taken or
derived from btsynth; see http://scottman.net/2012/btsynth
"""
from __future__ import print_function
import copy
import itertools
import random
import time

import numpy as np

from spec.form import GRSpec

# inline:
#
# import polytope
# import tulip.abstract.prop2part
# import matplotlib


class GridWorld(object):
    """4-connected grids with primitives like obstacles and goals."""

    def __init__(self, gw_desc=None, prefix="Y"):
        """Load gridworld described in given string, or make empty instance.

        @param gw_desc: String containing a gridworld description, or
                        None to create an empty instance.
        @param prefix: String to be used as prefix for naming
                       gridworld cell variables.
        """
        if gw_desc is not None:
            self.loads(gw_desc)
        else:
            self.W = None
            self.init_list = []
            self.goal_list = []
        self.prefix = prefix
        self.offset = (0, 0)

    def __eq__(self, other):
        """Test for equality.

        Does not compare prefixes of cell variable names.
        """
        if self.W is None and other.W is None:
            return True
        if self.W is None or other.W is None:
            return False  # Only one of the two is undefined.
        if self.size() != other.size():
            return False
        if np.all(self.W != other.W):
            return False
        if self.goal_list != other.goal_list:
            return False
        if self.init_list != other.init_list:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self.pretty(show_grid=True)

    def __getitem__(self, key, use_next=False, nonbool=True):
        """Return variable name corresponding to this cell.

        Supports negative wrapping, e.g., if Y is an instance of
        GridWorld, then Y[-1,-1] will return the variable name of the
        cell in the bottom-right corner, Y[0,-1] the name of the
        top-right corner cell, etc.  As usual in Python, you can only
        wrap around once.

        @param next: Use the primed (i.e., state at next time step)
                 form of the variable.
        @param nonbool: If True, then use variables with integer domains.
        """
        if self.W is None:
            raise ValueError("Gridworld is empty; no names available.")
        if len(key) != len(self.W.shape):
            raise ValueError("malformed gridworld key.")
        if (
                key[0] < -self.W.shape[0] or
                key[1] < -self.W.shape[1] or
                key[0] >= self.W.shape[0] or
                key[1] >= self.W.shape[1]):
            raise ValueError("gridworld key is out of bounds.")
        if key[0] < 0:
            key = (self.W.shape[0] + key[0], key[1])
        if key[1] < 0:
            key = (key[0], self.W.shape[1] + key[1])
        if nonbool:
            if use_next:
                return (
                    "((X (" + str(self.prefix) +
                    "_r = " + str(key[0] + self.offset[0]) +
                    ")) && (X (" + str(self.prefix) +
                    "_c = " + str(key[1] + self.offset[1]) + ")))")
            else:
                return (
                    "((" + str(self.prefix) +
                    "_r = " + str(key[0] + self.offset[0]) +
                    ") && (" + str(self.prefix) +
                    "_c = " + str(key[1] + self.offset[1]) + "))")
        else:
            if use_next:
                out = 'X '
            else:
                out = ''
            out += str(self.prefix) + "_" + \
                str(key[0] + self.offset[0]) + "_" + \
                str(key[1] + self.offset[1])
            return out

    def __copy__(self):
        return GridWorld(self.dumps(), prefix=self.prefix)

    def copy(self):
        """Deep-copy GridWorld instance.

        The copy has no dependence on the original.
        """
        return self.__copy__()

    def state(self, key, offset=(0, 0), nonbool=True):
        """Return dictionary form of state with keys of variable names.

        Supports negative indices for key, e.g., as in __getitem__.

        The offset argument is motivated by the use-case of multiple
        agents whose moves are governed by separate "gridworlds" but
        who interact in a shared space; with an offset, we can make
        "sub-gridworlds" and enforce rules like mutual exclusion.

        @param nonbool: If True, then use variables with integer domains.
        """
        if self.W is None:
            raise ValueError("Gridworld is empty; no cells exist.")
        if len(key) != len(self.W.shape):
            raise ValueError("malformed gridworld key.")
        if (
                key[0] < -self.W.shape[0] or
                key[1] < -self.W.shape[1] or
                key[0] >= self.W.shape[0] or
                key[1] >= self.W.shape[1]):
            raise ValueError("gridworld key is out of bounds.")
        if key[0] < 0:
            key = (self.W.shape[0] + key[0], key[1])
        if key[1] < 0:
            key = (key[0], self.W.shape[1] + key[1])
        output = dict()
        if nonbool:
            output[self.prefix + "_r"] = key[0] + offset[0]
            output[self.prefix + "_c"] = key[1] + offset[1]
        else:
            for i in range(self.W.shape[0]):
                for j in range(self.W.shape[1]):
                    output[self.prefix + "_" +
                           str(i + offset[0]) + "_" + str(j + offset[1])] = 0
            output[self.prefix + "_" +
                   str(key[0] + offset[0]) + "_" + str(key[1] + offset[1])] = 1
        return output

    def is_empty(self, coord, extend=False):
        """Is cell at coord empty?

        @param coord: (row, column) pair; supports negative indices.
        @param extend: If True, then do not wrap indices and treat any
                 cell outside the grid as being occupied.
        """
        if self.W is None:
            raise ValueError("Gridworld is empty; no cells exist.")
        if len(coord) != len(self.W.shape):
            raise ValueError("malformed gridworld coord.")
        if extend and (coord[0] < 0 or coord[1] < 0
                       or coord[0] > self.W.shape[0] - 1
                       or coord[1] > self.W.shape[1] - 1):
            return False
        if self.W[coord[0]][coord[1]] == 0:
            return True
        else:
            return False

    def mark_occupied(self, coord):
        """Mark cell at coord as statically (permanently) occupied."""
        if self.W is None:
            raise ValueError("Gridworld is empty; no cells exist.")
        self.W[coord[0]][coord[1]] = 1

    def mark_empty(self, coord):
        """Mark cell at coord as empty."""
        if self.W is None:
            raise ValueError("Gridworld is empty; no cells exist.")
        self.W[coord[0]][coord[1]] = 0

    def is_reachable(self, start, stop):
        """Decide whether there is a path from start cell to stop.

        Assume the gridworld is 4-connected.

        @param start: (row, column) pair; supports negative indices.
        @param stop: same as start argument.

        @return: True if there is a path, False otherwise.
        """
        # Check input values and handle negative coordinates
        if self.W is None:
            raise ValueError("Gridworld is empty; no names available.")

        if len(start) != len(self.W.shape):
            raise ValueError("malformed gridworld start coordinate.")
        if (
                start[0] < -self.W.shape[0] or
                start[1] < -self.W.shape[1] or
                start[0] >= self.W.shape[0] or
                start[1] >= self.W.shape[1]):
            raise ValueError("gridworld start coordinate is out of bounds.")
        if start[0] < 0:
            start = (self.W.shape[0] + start[0], start[1])
        if start[1] < 0:
            start = (start[0], self.W.shape[1] + start[1])

        if len(stop) != len(self.W.shape):
            raise ValueError("malformed gridworld stop coordinate.")
        if (
                stop[0] < -self.W.shape[0] or
                stop[1] < -self.W.shape[1] or
                stop[0] >= self.W.shape[0] or
                stop[1] >= self.W.shape[1]):
            raise ValueError("gridworld stop coordinate is out of bounds.")
        if stop[0] < 0:
            stop = (self.W.shape[0] + stop[0], stop[1])
        if stop[1] < 0:
            stop = (stop[0], self.W.shape[1] + stop[1])

        # Quick sanity check
        if not (self.is_empty(start) and self.is_empty(stop)):
            return False

        # Similar to depth-first search
        OPEN = [start]
        CLOSED = []
        while len(OPEN) > 0:
            current = OPEN.pop()
            if current == stop:
                return True
            for (i, j) in [(1, 0), (-1, 0), (0, 1), (0, -1)]:
                if (
                        current[0] + i < 0 or
                        current[0] + i >= self.W.shape[0] or
                        current[1] + j < 0 or
                        current[1] + j >= self.W.shape[1]):
                    continue
                if (
                        self.is_empty((current[0] + i, current[1] + j)) and
                        (current[0] + i, current[1] + j) not in CLOSED and
                        (current[0] + i, current[1] + j) not in OPEN):
                    OPEN.append((current[0] + i, current[1] + j))
            CLOSED.append(current)
        return False

    def plot(self, font_pt=18, show_grid=False, grid_width=2,
             troll_list=None, axes=None):
        """Draw figure depicting this gridworld.

        Figure legend (symbolic form in parenthesis):
          - "I" ('m+') : possible initial position;
          - "G" ('r*') : goal;
          - "E" ('gx') : goal of a troll; its extent is indicated by gray cells

        @param font_pt: size (in points) for rendering text in the
                 figure.  If 0, then use symbols instead (see legend above).
        @param troll_list: ...same as the argument with the same name
                 given to L{add_trolls}.

        @param axes: Instance of matplotlib.axes._subplots.AxesSubplot
                 on which to draw, or None, in which case a new figure
                 is created.
        """
        if troll_list is None:
            troll_list = list()
        try:
            import matplotlib.cm as mpl_cm
            import matplotlib.patches
            import matplotlib.pyplot as plt
        except ImportError:
            print('matplotlib not available, so skipping GridWorld.plot()')
            return

        W = self.W.copy()
        W = np.ones(shape=W.shape) - W
        if axes is None:
            fig = plt.figure()
            ax = plt.subplot(111)
        else:
            ax = axes
        ax.imshow(W, cmap=mpl_cm.gray,
                  aspect="equal", interpolation="nearest",
                  vmin=0., vmax=1.)
        xmin, xmax, ymin, ymax = plt.axis()
        x_steps = np.linspace(xmin, xmax, W.shape[1] + 1)
        y_steps = np.linspace(ymin, ymax, W.shape[0] + 1)
        if show_grid:
            for k in x_steps:
                ax.plot([k, k], [ymin, ymax], 'k-', linewidth=grid_width)
            for k in y_steps:
                ax.plot([xmin, xmax], [k, k], 'k-', linewidth=grid_width)
            ax.axis([xmin, xmax, ymin, ymax])
        for p in self.init_list:
            if font_pt > 0:
                ax.text(p[1], p[0], "I", size=font_pt)
            else:
                ax.plot(p[1], p[0], 'm+')
        for p in self.goal_list:
            if font_pt > 0:
                ax.text(p[1], p[0], "G", size=font_pt)
            else:
                ax.plot(p[1], p[0], 'r*')
        for (center, radius) in troll_list:
            if font_pt > 0:
                ax.text(center[1], center[0], "E", size=font_pt)
            else:
                ax.plot(center[1], center[0], 'gx')
            if (
                    center[0] >= W.shape[0] or
                    center[0] < 0 or
                    center[1] >= W.shape[1] or
                    center[1] < 0):
                raise ValueError("troll center is outside of gridworld")
            t_offset = (max(0, center[0] - radius), max(0, center[1] - radius))
            t_size = [center[0] - t_offset[0] + radius +
                      1, center[1] - t_offset[1] + radius + 1]
            if t_offset[0] + t_size[0] >= W.shape[0]:
                t_size[0] = W.shape[0] - t_offset[0]
            if t_offset[1] + t_size[1] >= W.shape[1]:
                t_size[1] = W.shape[1] - t_offset[1]
            t_size = (t_size[0], t_size[1])
            for i in range(t_size[0]):
                for j in range(t_size[1]):
                    if self.W[i + t_offset[0]][j + t_offset[1]] == 0:
                        ax.add_patch(
                            matplotlib.patches.Rectangle(
                                (x_steps[
                                    j + t_offset[1]],
                                    y_steps[W.shape[0] - (i + t_offset[0])]),
                                1, 1, color=(.8, .8, .8)))
        ax.axis([xmin, xmax, ymin, ymax])

    def pretty(self, show_grid=False, line_prefix="",
               path=None, goal_order=False):
        """Return pretty-for-printing string.

        @param show_grid: If True, then grid the pretty world and show
                          row and column labels along the outer edges.
        @param line_prefix: prefix each line with this string.
        """
        if path is None:
            path = list()
        compress = lambda p: [p[n]
                              for n in range(len(p) - 1) if p[n] != p[n + 1]]
        # See comments in code for the method loads regarding values in W
        if self.W is None:
            return ""

        # LEGEND:
        #  * - wall (as used in original world matrix definition);
        #  G - goal location;
        #  I - possible initial location.
        out_str = line_prefix

        def direct(c1, c2):
            (y1, x1) = c1
            (y2, x2) = c2
            if x1 > x2:
                return "<"
            elif x1 < x2:
                return ">"
            elif y1 > y2:
                return "^"
            elif y1 < y2:
                return "v"
            else:  # c1 == c2
                return "."
        if show_grid:
            out_str += "  " + "".join([str(k).rjust(2)
                                       for k in range(self.W.shape[1])]) + "\n"
        else:
            out_str += "-" * (self.W.shape[1] + 2) + "\n"
        # if path:
        #    path = compress(path)
        for i in range(self.W.shape[0]):
            out_str += line_prefix
            if show_grid:
                out_str += "  " + "-" * (self.W.shape[1] * 2 + 1) + "\n"
                out_str += line_prefix
                out_str += str(i).rjust(2)
            else:
                out_str += "|"
            for j in range(self.W.shape[1]):
                if show_grid:
                    out_str += "|"
                if self.W[i][j] == 0:
                    if (i, j) in self.init_list:
                        out_str += "I"
                    elif (i, j) in self.goal_list:
                        if goal_order:
                            out_str += str(self.goal_list.index((i, j)))
                        else:
                            out_str += "G"
                    elif (i, j) in path:
                        indices = (n for (n, c) in enumerate(
                            path) if c == (i, j))
                        for x in indices:
                            d = direct((i, j), path[(x + 1) % len(path)])
                            if d != ".":
                                break
                        out_str += d
                    else:
                        out_str += " "
                elif self.W[i][j] == 1:
                    out_str += "*"
                else:
                    raise ValueError("Unrecognized internal world W encoding.")
            out_str += "|\n"
        out_str += line_prefix
        if show_grid:
            out_str += "  " + "-" * (self.W.shape[1] * 2 + 1) + "\n"
        else:
            out_str += "-" * (self.W.shape[1] + 2) + "\n"
        return out_str

    def size(self):
        """Return size of gridworld as a tuple in row-major order."""
        if self.W is None:
            return (0, 0)
        else:
            return self.W.shape

    def loads(self, gw_desc):
        """Reincarnate using given gridworld description string.

        @param gw_desc: String containing a gridworld description.

        In a gridworld description, any line beginning with # is
        ignored (regarded as a comment). The first non-blank and
        non-comment line must give the grid size as two positive
        integers separated by whitespace, with the first being the
        number of rows and the second the number of columns.

        Each line after the size line is used to construct a row of
        the gridworld. These are read in order with maximum number of
        lines being the number of rows in the gridworld.  A row
        definition is whitespace-sensitive up to the number of columns
        (any characters beyond the column count are ignored, so in
        particular trailing whitespace is allowed) and can include the
        following symbols:

          - ' ' : an empty cell,
          - '*' : a statically occupied cell,
          - 'I' : possible initial cell,
          - 'G' : goal cell (must be visited infinitely often).

        If the end of file is reached before all rows have been
        constructed, then the remaining rows are assumed to be empty.
        After all rows have been constructed, the remainder of the
        file is ignored.
        """
        ###################################################
        # Internal format notes:
        #
        # W is a matrix of integers with the same shape as the
        # gridworld.  Each element has value indicating properties of
        # the corresponding cell, according the following key.
        #
        # 0 - empty,
        # 1 - statically (permanently) occupied.
        ###################################################
        W = None
        init_list = []
        goal_list = []
        row_index = -1
        for line in gw_desc.splitlines():
            if row_index != -1:
                # Size has been read, so we are processing row definitions
                if row_index >= W.shape[0]:
                    break
                for j in range(min(len(line), W.shape[1])):
                    if line[j] == " ":
                        W[row_index][j] = 0
                    elif line[j] == "*":
                        W[row_index][j] = 1
                    elif line[j] == "I":
                        init_list.append((row_index, j))
                    elif line[j] == "G":
                        goal_list.append((row_index, j))
                    else:
                        raise ValueError(
                            "unrecognized row symbol \"" +
                            str(line[j]) + "\".")
                row_index += 1
            else:
                # Still looking for gridworld size in the given string
                if len(line.strip()) == 0 or line.lstrip()[0] == "#":
                    continue  # Ignore blank and comment lines
                line_el = line.split()
                W = np.zeros((int(line_el[0]), int(line_el[1])),
                             dtype=np.int32)
                row_index = 0

        if W is None:
            raise ValueError("malformed gridworld description.")

        # Arrived here without errors, so actually reincarnate
        self.W = W
        self.init_list = init_list
        self.goal_list = goal_list

    def load(self, gw_file):
        """Read description from given file.

        Merely a convenience wrapper for the L{loads} method.
        """
        with open(gw_file, "r") as f:
            self.loads(f.read())

    def dumps(self, line_prefix=""):
        """Dump gridworld description string.

        @param line_prefix: prefix each line with this string.
        """
        if self.W is None:
            raise ValueError("Gridworld does not exist.")
        out_str = line_prefix + " ".join([str(i) for i in self.W.shape]) + "\n"
        for i in range(self.W.shape[0]):
            out_str += line_prefix
            for j in range(self.W.shape[1]):
                if self.W[i][j] == 0:
                    if (i, j) in self.init_list:
                        out_str += "I"
                    elif (i, j) in self.goal_list:
                        out_str += "G"
                    else:
                        out_str += " "
                elif self.W[i][j] == 1:
                    out_str += "*"
                else:
                    raise ValueError("Unrecognized internal world W encoding.")
            out_str += "\n"
        return out_str

    def dump_subworld(self, size, offset=(0, 0), prefix="Y", extend=False):
        """Generate new GridWorld instance from part of current one.

        Does not perform automatic truncation (to make desired
        subworld fit); instead a ValueError exception is raised.
        However, the "extend" argument can be used to achieve
        something similar.

        Possible initial positions and goals are not included in the
        returned GridWorld instance.

        @param size: (height, width)
        @param prefix: String to be used as prefix for naming
                 subgridworld cell variables.
        @param extend: If True, then any size and offset is permitted,
                 where any positions outside the actual gridworld are
                 assumed to be occupied.

        @rtype: L{GridWorld}
        """
        if self.W is None:
            raise ValueError("Gridworld does not exist.")
        if len(size) != len(self.W.shape) or len(offset) != len(self.W.shape):
            raise ValueError("malformed size or offset.")
        if not extend:
            if (
                    offset[0] < 0 or
                    offset[0] >= self.W.shape[0] or
                    offset[1] < 0 or
                    offset[1] >= self.W.shape[1]):
                raise ValueError("offset is out of bounds.")
            if (
                    size[0] < 1 or
                    size[1] < 1 or
                    offset[0] + size[0] > self.W.shape[0] or
                    offset[1] + size[1] > self.W.shape[1]):
                raise ValueError("unworkable subworld size, given offset.")
            sub = GridWorld(prefix=prefix)
            sub.W = self.W[offset[0]:(
                offset[0] + size[0]), offset[1]:(offset[1] + size[1])].copy()
        else:
            sub = GridWorld(prefix=prefix)
            sub.W = np.ones(size)
            self_offset = (max(offset[0], 0), max(offset[1], 0))
            self_offset = (min(self_offset[0], self.W.shape[
                           0] - 1), min(self_offset[1], self.W.shape[1] - 1))
            sub_offset = (max(-offset[0], 0), max(-offset[1], 0))
            sub_offset = (min(sub_offset[0], sub.W.shape[
                          0] - 1), min(sub_offset[1], sub.W.shape[1] - 1))
            actual_size = (
                min(size[0],
                    self.W.shape[0] - self_offset[0],
                    sub.W.shape[0] - sub_offset[0]),
                min(size[1],
                    self.W.shape[1] - self_offset[1],
                    sub.W.shape[1] - sub_offset[1]))
            sub.W[sub_offset[0]:(sub_offset[0] + actual_size[0]),
                  sub_offset[1]:(sub_offset[1] + actual_size[1])] = self.W[
                self_offset[0]:(self_offset[0] + actual_size[0]),
                self_offset[1]:(self_offset[1] + actual_size[1])]
        return sub

    def dump_ppartition(self, side_lengths=(1., 1.),
                        offset=(0., 0.), nonbool=True):
        """Return proposition-preserving partition from this gridworld.

        Adjacency of cells is as returned by prop2partition.prop2part().

        @param side_lengths: pair (W, H) giving width and height of
                 each cell, assumed to be the same across the grid.
        @param offset: 2-dimensional coordinate declaring where the
                 bottom-left corner of the gridworld should be placed
                 in the continuous space; default places it at the origin.

        @rtype: L{PropPreservingPartition<prop2part.PropPreservingPartition>}
        """
        try:
            from polytope import Polytope
        except ImportError:
            raise ImportError('GridWorld.dump_ppartition() requires '
                              'the Python package polytope.')
        try:
            from tulip.abstract import prop2partition
        except ImportError:
            raise ImportError('`GridWorld.dump_ppartition()` requires '
                              'the module `tulip.abstract`.')

        if self.W is None:
            raise ValueError("Gridworld does not exist.")
        domain = Polytope(A=np.array([[0, -1],
                                      [0, 1],
                                      [-1, 0],
                                      [1, 0]], dtype=np.float64),
                          b=np.array([-offset[1],
                                      offset[1] + self.W.shape[0] *
                                      side_lengths[1],
                                      -offset[0],
                                      offset[0] +
                                      self.W.shape[1] * side_lengths[0]],
                                     dtype=np.float64))
        cells = {}
        for i in range(self.W.shape[0]):
            for j in range(self.W.shape[1]):
                if nonbool:
                    cell_var = self.__getitem__((i, j))
                else:
                    cell_var = self.prefix + "_" + str(i) + "_" + str(j)
                # adjacency[i]
                cells[cell_var] \
                    = Polytope(
                        A=np.array(
                            [[0, -1],
                             [0, 1],
                             [-1, 0],
                             [1, 0]], dtype=np.float64),
                        b=np.array(
                            [-offset[1] -
                             (self.W.shape[0] - i - 1) * side_lengths[1],
                             offset[1] +
                             (self.W.shape[0] - i) * side_lengths[1],
                             -offset[0] - j * side_lengths[0],
                             offset[0] + (j + 1) * side_lengths[0]],
                            dtype=np.float64))
        return prop2partition.prop2part(domain, cells)

    def spec(self, offset=(0, 0), controlled_dyn=True, nonbool=True):
        """Return GRSpec instance describing this gridworld.

        The offset argument is motivated by the use-case of multiple
        agents whose moves are governed by separate "gridworlds" but
        who interact in a shared space; with an offset, we can make
        "sub-gridworlds" and enforce rules like mutual exclusion.

        If nonbool is False, then variables are named according to
        prefix_R_C, where prefix is given (attribute of this GridWorld
        object), R is the row, and C is the column of the cell
        (0-indexed).  If nonbool is True (default), cells are
        identified with subformulae of the form::

          ((prefix_r = R) & (prefix_c = C))

        L{GridWorld.__getitem__} and L{extract_coord} provide
        reference implementations.

        @param offset: index offset to apply when generating the
                 specification; e.g., given prefix of "Y",
                 offset=(2,1) would cause the variable for the cell at
                 (0,3) to be named Y_2_4.

        @param controlled_dyn: whether to treat this gridworld as
                 describing controlled ("system") or uncontrolled
                 ("environment") variables.

        @param nonbool: If True, then use variables with integer domains.

        @rtype: L{GRSpec}
        """
        if self.W is None:
            raise ValueError("Gridworld does not exist.")
        row_low = 0
        row_high = self.W.shape[0]
        col_low = 0
        col_high = self.W.shape[1]
        spec_trans = []
        orig_offset = copy.copy(self.offset)
        if nonbool:
            self.offset = (0, 0)
        else:
            self.offset = offset
        # Safety, transitions
        for i in range(row_low, row_high):
            for j in range(col_low, col_high):
                if self.W[i][j] == 1:
                    continue  # Cannot start from an occupied cell.
                spec_trans.append(self.__getitem__(
                    (i, j), nonbool=nonbool) + " -> (")
                # Normal transitions:
                spec_trans[-1] += self.__getitem__((i, j),
                                                   use_next=True,
                                                   nonbool=nonbool)
                if i > row_low and self.W[i - 1][j] == 0:
                    spec_trans[-1] += " || " + \
                        self.__getitem__(
                            (i - 1, j), use_next=True, nonbool=nonbool)
                if j > col_low and self.W[i][j - 1] == 0:
                    spec_trans[-1] += " || " + \
                        self.__getitem__(
                            (i, j - 1), use_next=True, nonbool=nonbool)
                if i < row_high - 1 and self.W[i + 1][j] == 0:
                    spec_trans[-1] += " || " + \
                        self.__getitem__(
                            (i + 1, j), use_next=True, nonbool=nonbool)
                if j < col_high - 1 and self.W[i][j + 1] == 0:
                    spec_trans[-1] += " || " + \
                        self.__getitem__(
                            (i, j + 1), use_next=True, nonbool=nonbool)
                spec_trans[-1] += ")"

        # Safety, static
        for i in range(row_low, row_high):
            for j in range(col_low, col_high):
                if self.W[i][j] == 1:
                    spec_trans.append(
                        "!(" + self.__getitem__((i, j), use_next=True,
                                                nonbool=nonbool) + ")")

        # Safety, mutex; only needed when using boolean variables for cells
        if not nonbool:
            pos_indices = [k for k in itertools.product(
                range(row_low, row_high), range(col_low, col_high))]
            disj = []
            for outer_ind in pos_indices:
                conj = []
                if (
                        outer_ind != (-1, -1) and
                        self.W[outer_ind[0]][outer_ind[1]] == 1):
                    continue
                if outer_ind == (-1, -1):
                    conj.append(self.prefix + "_n_n'")
                else:
                    conj.append(
                        self.__getitem__(
                            (outer_ind[0], outer_ind[1]),
                            use_next=True, nonbool=nonbool))
                for inner_ind in pos_indices:
                    if (
                            (
                                inner_ind != (-1, -1) and
                                self.W[inner_ind[0]][inner_ind[1]] == 1) or
                            outer_ind == inner_ind):
                        continue
                    if inner_ind == (-1, -1):
                        conj.append("(!X " + self.prefix + "_n_n)")
                    else:
                        conj.append(
                            "(!" +
                            self.__getitem__(
                                (inner_ind[0], inner_ind[1]),
                                use_next=True,
                                nonbool=nonbool) + ")")
                disj.append("(" + " && ".join(conj) + ")")
            spec_trans.append("\n|| ".join(disj))

        if nonbool:
            sys_vars = {self.prefix + "_r": (0, self.W.shape[0] - 1),
                        self.prefix + "_c": (0, self.W.shape[1] - 1)}
        else:
            sys_vars = set()
            for i in range(row_low, row_high):
                for j in range(col_low, col_high):
                    sys_vars.add(self.__getitem__((i, j), nonbool=nonbool))

        if nonbool:
            initspec = [self.__getitem__(loc, nonbool=nonbool)
                        for loc in self.init_list]
        else:
            initspec = []
            for loc in self.init_list:
                mutex = [self.__getitem__((loc[0], loc[1]), nonbool=nonbool)]
                mutex.extend(["!" + ovar for ovar in sys_vars if ovar !=
                              self.__getitem__(loc, nonbool=nonbool)])
                initspec.append("(" + " && ".join(mutex) + ")")
        init_str = " || ".join(initspec)

        spec_goal = []
        for loc in self.goal_list:
            spec_goal.append(self.__getitem__(loc, nonbool=nonbool))

        self.offset = orig_offset
        if controlled_dyn:
            return GRSpec(sys_vars=sys_vars,
                          sys_init=init_str,
                          sys_safety=spec_trans,
                          sys_prog=spec_goal)
        else:
            return GRSpec(env_vars=sys_vars,
                          env_init=init_str,
                          env_safety=spec_trans,
                          env_prog=spec_goal)

    def scale(self, xf=1, yf=1):
        """Return a new gridworld equivalent to this but scaled by integer
        factor (xf, yf). In the new world, obstacles are increased in size but
        initials and goals change their position only. If this world is of size
        (h, w) then the returned world will be of size (h*yf, w*xf).

        @param xf: integer scaling factor for columns
        @param yf: integer scaling factor for rows

        @rtype: L{GridWorld}
        """
        shape_scaled = (self.W.shape[0] * yf, self.W.shape[1] * xf)
        scaleW = np.zeros(shape_scaled, dtype=np.int32)
        scale_goal = []
        scale_init = []
        for row in range(shape_scaled[0]):
            for col in range(shape_scaled[1]):
                (y, x) = (row / yf, col / xf)
                (yr, xr) = (row % yf, col % xf)
                if self.W[y, x] == 1:
                    scaleW[row, col] = 1
                if (yr, xr) == (0, 0):
                    if (y, x) in self.goal_list:
                        scale_goal.append((row, col))
                    if (y, x) in self.init_list:
                        scale_init.append((row, col))
        scale_gw = GridWorld(prefix=self.prefix)
        scale_gw.W = scaleW
        scale_gw.goal_list = scale_goal
        scale_gw.init_list = scale_init
        return scale_gw


def random_world(size, wall_density=.2, num_init=1, num_goals=2, prefix="Y",
                 ensure_feasible=False, timeout=None,
                 num_trolls=0):
    """Generate random gridworld of given size.

    While an instance of GridWorld is returned, other views of the
    result are possible; e.g., to obtain a description string, use
    L{GridWorld.dumps}.

    @param size: a pair, indicating number of rows and columns.
    @param wall_density: the ratio of walls to total number of cells.
    @param num_init: number of possible initial positions.
    @param num_goals: number of positions to be visited infinitely often.
    @param prefix: string to be used as prefix for naming gridworld
             cell variables.

    @param num_trolls: number of random trolls to generate, each
             occupies an area of radius 1.  If nonzero, then a list
             specifying the trolls will also be returned.

    @param ensure_feasible: guarantee that all goals and initial
             positions are mutually reachable, assuming a 4-connected
             grid. This method may not be complete, i.e., may fail to
             return a feasible random gridworld with the given
             parameters.  Note that "feasibility" does not account for
             nondeterminism (in particular, nonzero num_trolls
             argument has no effect.)

    @param timeout: if ensure_feasible, then quit if no correct random
             world is found before timeout seconds.  If timeout is
             None (default), then do not impose time constraints.

    @rtype: L{GridWorld}, or None if timeout occurs.

    """
    if ensure_feasible and timeout is not None:
        st = time.time()
    num_cells = size[0] * size[1]
    goal_list = []
    init_list = []
    troll_list = []
    W = np.zeros(num_cells, dtype=np.int32)
    num_blocks = int(np.round(wall_density * num_cells))
    for i in range(num_goals):
        avail_inds = np.array(range(num_cells))[W == 0]
        avail_inds = [k for k in avail_inds if k not in goal_list]
        goal_list.append(
            avail_inds[np.random.randint(low=0, high=len(avail_inds))])
    for i in range(num_init):
        avail_inds = np.array(range(num_cells))[W == 0]
        avail_inds = [
            k for k in avail_inds if k not in goal_list and k not in init_list]
        init_list.append(
            avail_inds[np.random.randint(low=0, high=len(avail_inds))])
    for i in range(num_trolls):
        avail_inds = np.array(range(num_cells))[W == 0]
        avail_inds = [
            k for k in avail_inds
            if (
                k not in goal_list and
                k not in init_list and
                k not in troll_list)]
        troll_list.append(
            avail_inds[np.random.randint(low=0, high=len(avail_inds))])
    bcounter = 0
    while bcounter < num_blocks:  # Add blocks (or "wall cells")
        avail_inds = np.array(range(num_cells))[W == 0]
        avail_inds = [
            k for k in avail_inds
            if (
                k not in goal_list and
                k not in init_list and
                k not in troll_list)]
        changed_index = np.random.randint(low=0, high=len(avail_inds))
        W[avail_inds[changed_index]] = 1
        bcounter += 1
        if ensure_feasible:
            if (timeout is not None) and (time.time() - st > timeout):
                return None
            # If feasibility must be guaranteed, then check whether
            # the newly unreachable cell is permissible.
            W_tmp = W.reshape(size)
            goal_list_tmp = [(k / size[1], k % size[1]) for k in goal_list]
            init_list_tmp = [(k / size[1], k % size[1]) for k in init_list]
            troll_list_tmp = [(k / size[1], k % size[1]) for k in troll_list]
            world = GridWorld(prefix=prefix)
            world.W = W_tmp
            chain_of_points = init_list_tmp[:]
            chain_of_points.extend(goal_list_tmp)
            is_feasible = True
            for i in range(len(chain_of_points)):
                if not world.is_reachable(
                        chain_of_points[i],
                        chain_of_points[(i + 1) % len(chain_of_points)]):
                    is_feasible = False
                    break
            if not is_feasible:
                W[avail_inds[changed_index]] = 0
                bcounter -= 1
    # Reshape the gridworld to final form; build and return the result.
    W = W.reshape(size)
    goal_list = [(k / size[1], k % size[1]) for k in goal_list]
    init_list = [(k / size[1], k % size[1]) for k in init_list]
    troll_list = [((k / size[1], k % size[1]), 1) for k in troll_list]
    world = GridWorld(prefix=prefix)
    world.W = W
    world.goal_list = goal_list
    world.init_list = init_list

    if num_trolls > 0:
        return world, troll_list
    else:
        return world


def narrow_passage(size, passage_width=1, num_init=1, num_goals=2,
                   passage_length=0.4, ptop=None, prefix="Y"):
    """Generate a narrow-passage world: this is a world containing
    two zones (initial, final) with a tube connecting them.

    @param size: a pair, indicating number of rows and columns.
    @param passage_width: the width of the connecting passage in cells.
    @param passage_length: the length of the passage as a proportion of the
                           width of the world.
    @param num_init: number of possible initial positions.
    @param num_goals: number of positions to be visited infinitely often.
    @param ptop: row number of top of passage, default (None) is random
    @param prefix: string to be used as prefix for naming gridworld
                   cell variables.

    @rtype: L{GridWorld}
    """
    (w, h) = size
    if w < 3 or h < 3:
        raise ValueError("Gridworld too small: minimum dimension 3")
    Z = unoccupied(size, prefix)
    # Zone width is 30% of world width by default
    zone_width = ((1.0 - passage_length) / 2.0) * size[1]
    izone = int(max(1, zone_width))  # boundary of left zone
    gzone = size[1] - int(max(1, zone_width))  # boundary of right zone
    if izone * size[0] < num_init or gzone * size[0] < num_goals:
        raise ValueError("Too many initials/goals for grid size")
    if ptop is None:
        ptop = np.random.randint(0, size[0] - passage_width)
    passage = range(ptop, ptop + passage_width)
    for y in range(0, size[0]):
        if y not in passage:
            for x in range(izone, gzone):
                Z.W[y][x] = 1
    avail_cells = [(y, x) for y in range(size[0]) for x in range(izone)]
    Z.init_list = random.sample(avail_cells, num_init)
    avail_cells = [(y, x) for y in range(size[0])
                   for x in range(gzone, size[1])]
    Z.goal_list = random.sample(avail_cells, num_goals)
    return Z


def unoccupied(size, prefix="Y"):
    """Generate entirely unoccupied gridworld of given size.

    @param size: a pair, indicating number of rows and columns.
    @param prefix: String to be used as prefix for naming gridworld
                   cell variables.
    @rtype: L{GridWorld}
    """
    if len(size) < 2:
        raise TypeError("invalid gridworld size.")
    return GridWorld(str(size[0]) + " " + str(size[1]), prefix="Y")


def add_trolls(Y, troll_list, prefix="X", start_anywhere=False, nonbool=True,
               get_moves_lists=True):
    """Create GR(1) specification with troll-like obstacles.

    Trolls are introduced into the specification with names derived
    from the given prefix and a number (matching the order in
    troll_list).  Note that mutual exclusion is only between the
    controlled "Y gridworld" position and each troll, but not
    between trolls.

    @type Y: L{GridWorld}
    @param Y: The controlled gridworld, describing in particular
             static obstacles that must be respected by the trolls.

    @param troll_list: List of pairs of center position, to which the
             troll must always eventually return, and radius defining
             the extent of the trollspace.  The radius is measured
             using infinity-norm.
    @param start_anywhere: If True, then initial troll position can be
             anywhere in its trollspace.  Else (default), the troll is
             assumed to begin each game at its center position.
    @param nonbool: If True, then use variables with integer domains.
    @param get_moves_lists: Consult returned value description below.

    @rtype: (L{GRSpec}, list)

    @return: If get_moves_lists is True, then returns (spec, moves_N)
             where spec is the specification incorporating all of the
             trolls, and moves_N is a list of lists of states (where
             "state" is given as a dictionary with keys of variable
             names), where the length of moves_N is equal to the
             number of trolls, and each element of moves_N is a list
             of possible states of that the corresponding troll
             (dynamic obstacle).  If get_moves_lists is False, then
             moves_N is not returned and not computed.
    """
    X = []
    X_ID = -1
    if get_moves_lists:
        moves_N = []
    (num_rows, num_cols) = Y.size()
    for (center, radius) in troll_list:
        if (
                center[0] >= num_rows or
                center[0] < 0 or
                center[1] >= num_cols or
                center[1] < 0):
            raise ValueError("troll center is outside of gridworld")
        t_offset = (max(0, center[0] - radius), max(0, center[1] - radius))
        t_size = [center[0] - t_offset[0] + radius +
                  1, center[1] - t_offset[1] + radius + 1]
        if t_offset[0] + t_size[0] >= num_rows:
            t_size[0] = num_rows - t_offset[0]
        if t_offset[1] + t_size[1] >= num_cols:
            t_size[1] = num_cols - t_offset[1]
        t_size = (t_size[0], t_size[1])
        X_ID += 1
        X.append((t_offset, Y.dump_subworld(
            t_size, offset=t_offset, prefix=prefix + "_" + str(X_ID))))
        X[-1][1].goal_list = [(center[0] - t_offset[0],
                               center[1] - t_offset[1])]
        if start_anywhere:
            X[-1][1].init_list = []
            for i in range(X[-1][1].size()[0]):
                for j in range(X[-1][1].size()[1]):
                    if X[-1][1].is_empty((i, j)):
                        X[-1][1].init_list.append((i, j))
        else:
            X[-1][1].init_list = [(center[0] - t_offset[0],
                                   center[1] - t_offset[1])]
        if get_moves_lists:
            moves_N.append([])
            for i in range(t_size[0]):
                for j in range(t_size[1]):
                    moves_N[-1].append(
                        X[-1][1].state((i, j),
                        offset=t_offset, nonbool=nonbool))

    spec = Y.spec(controlled_dyn=True, nonbool=nonbool)
    for Xi in X:
        spec |= Xi[1].spec(offset=(-Xi[0][0], -Xi[0][1]),
                           controlled_dyn=False, nonbool=nonbool)

    # Mutual exclusion
    for i in range(Y.size()[0]):
        for j in range(Y.size()[1]):
            for Xi in X:
                if (
                        i >= Xi[0][0] and
                        i < Xi[0][0] + Xi[1].size()[0] and
                        j >= Xi[0][1] and
                        j < Xi[0][1] + Xi[1].size()[1]):
                    if nonbool:
                        Xivar = (
                            "((X " + Xi[1].prefix +
                            "_r = " + str(i - Xi[0][0]) +
                            ") & (X " + Xi[1].prefix + "_c = " +
                            str(j - Xi[0][1]) + "))")
                    else:
                        Xivar = (
                            "X " + Xi[1].prefix +
                            "_" + str(i) + "_" + str(j))
                    spec.sys_safety.append(
                        "!(" + Y.__getitem__(
                            (i, j), nonbool=nonbool, use_next=True) +
                        " && " + Xivar + ")")

    if get_moves_lists:
        return (spec, moves_N)
    return spec


def extract_coord(var_name):
    """Assuming prefix_R_C format, return (prefix,row,column) tuple.

    prefix is of type string, row and column are integers.

    The "nowhere" coordinate has form prefix_n_n. To indicate this,
    (-1, -1) is returned as the row, column position.

    If error, return None or throw exception.
    """
    if not isinstance(var_name, str):
        raise TypeError(
            "extract_coord: invalid argument type; must be string.")
    name_frags = var_name.split("_")
    if len(name_frags) < 3:
        return None
    try:
        if name_frags[-1] == "n" and name_frags[-2] == "n":
            # Special "nowhere" case
            return ("_".join(name_frags[:-2]), -1, -1)
        col = int(name_frags[-1])
        row = int(name_frags[-2])
    except ValueError:
        return None
    return ("_".join(name_frags[:-2]), row, col)


def animate_paths(Z, paths, jitter=0.0, save_prefix=None):
    """Animate a list of paths simultaneously in world Z using matplotlib.

    @param Z: Gridworld for which paths were generated.

    @param paths: List of paths to animate (one per robot). Each path
             is a list of pairs of gridworld cells, i.e., of the form
             [(r0, c0), (r1, c1), ...], where the first position is
             row r0 and column c0, etc.

    @param jitter: Random jitter added to each coordinate value in
             animation.  Makes the robot's path more visible by
             avoiding overlap.

    @param save_prefix: If not None, do not show an animation but
             produce a series of images "<save_prefix>nnn.png" which
             can be compiled into an animated GIF.
    """
    try:
        import matplotlib.animation
        import matplotlib.pyplot as plt
    except ImportError:
        print('matplotlib not available, '
              'so skipping gridworld.animate_paths()')
        return

    colors = 'rgbcmyk'
    fig = plt.figure()
    ax = fig.add_subplot(111)
    Z.plot(font_pt=min(288 / Z.W.shape[1], 48), show_grid=True, axes=ax)

    def update_line(num, dlist, lines):
        for (p, t), d in zip(lines, dlist):
            t.set_data(d[..., :num + 1])
            p.set_data(d[..., num])
        if save_prefix:
            fig.savefig(save_prefix + "%03d.png" % num)
        return lines,

    data = []
    lines = []
    for n, path in enumerate(paths):
        arr = np.array([[x, y] for (y, x) in path]).transpose()
        arr = np.add(arr, jitter * (np.random.rand(*arr.shape) - 0.5))
        data.append(arr)
        l, = ax.plot([], [], 'o', color=colors[n], markersize=10.0, zorder=2)
        l_trail, = ax.plot([], [], '-', color=colors[n], zorder=1)
        lines.append((l, l_trail))

    if not save_prefix:
        ani = matplotlib.animation.FuncAnimation(fig, update_line,
                                                 len(paths[0]),
                                                 fargs=(data, lines),
                                                 interval=500)
        plt.show()
    else:
        print("Writing %s000.png - %s%03d.png" %
              (save_prefix, save_prefix, len(paths[0])-1))
        for n in range(len(paths[0])):
            update_line(n, data, lines)



# Copyright by California Institute of Technology
# All rights reserved. See LICENSE file at:
# https://github.com/tulip-control/tulip-control
"""Interface to `omega` package.

`omega` constructs symbolic transducers,
represented as binary decision diagrams.
This module applies enumeration,
to return enumerated transducers.

U{https://pypi.python.org/pypi/omega}
"""
from __future__ import absolute_import
import logging
import time

try:
    import dd.bdd as _bdd
except ImportError:
    _bdd = None
try:
    from dd import cudd
except ImportError:
    cudd = None
try:
    import omega
    from omega.logic import bitvector as bv
    from omega.games import gr1
    from omega.symbolic import symbolic as sym
    from omega.games import enumeration as enum
except ImportError:
    omega = None
import networkx as nx


log = logging.getLogger(__name__)


def is_realizable(spec, use_cudd=False):
    """Return `True` if, and only if, realizable.

    See `synthesize_enumerated_streett` for more details.
    """
    aut = _grspec_to_automaton(spec)
    sym.fill_blanks(aut)
    bdd = _init_bdd(use_cudd)
    aut.bdd = bdd
    a = aut.build()
    t0 = time.time()
    z, _, _ = gr1.solve_streett_game(a)
    t1 = time.time()
    return gr1.is_realizable(z, a)


def synthesize_enumerated_streett(spec, use_cudd=False):
    """Return transducer enumerated as a graph.

    @type spec: `tulip.spec.form.GRSpec`
    @param use_cudd: efficient BDD computations with `dd.cudd`
    @rtype: `networkx.DiGraph`
    """
    aut = _grspec_to_automaton(spec)
    sym.fill_blanks(aut)
    bdd = _init_bdd(use_cudd)
    aut.bdd = bdd
    a = aut.build()
    assert a.action['sys'][0] != bdd.false
    t0 = time.time()
    z, yij, xijk = gr1.solve_streett_game(a)
    t1 = time.time()
    # unrealizable ?
    if not gr1.is_realizable(z, a):
        print('WARNING: unrealizable')
        return None
    t = gr1.make_streett_transducer(z, yij, xijk, a)
    t2 = time.time()
    (u,) = t.action['sys']
    assert u != bdd.false
    g = enum.action_to_steps(t, qinit=spec.qinit)
    h = _strategy_to_state_annotated(g, a)
    del u, yij, xijk
    t3 = time.time()
    log.info((
        'Winning set computed in {win} sec.\n'
        'Symbolic strategy computed in {sym} sec.\n'
        'Strategy enumerated in {enu} sec.').format(
            win=t1 - t0,
            sym=t2 - t1,
            enu=t3 - t2))
    return h


def is_circular(spec, use_cudd=False):
    """Return `True` if trivial winning set non-empty.

    @type spec: `tulip.spec.form.GRSpec`
    @param use_cudd: efficient BDD computations with `dd.cudd`
    @rtype: `bool`
    """
    aut = _grspec_to_automaton(spec)
    sym.fill_blanks(aut)
    bdd = _init_bdd(use_cudd)
    aut.bdd = bdd
    triv, t = gr1.trivial_winning_set(aut)
    return triv != t.bdd.false


def _init_bdd(use_cudd):
    if _bdd is None:
        raise ImportError(
            'Failed to import `dd.bdd`.\n'
            'Install package `dd`.')
    if not use_cudd:
        return _bdd.BDD()
    if cudd is None:
        raise ImportError(
            'Failed to import module `dd.cudd`.\n'
            'Compile the Cython bindings of `dd` to CUDD.')
    return cudd.BDD()


def _int_bounds(aut):
    """Create care set for enumeration.

    @type aut: `omega.symbolic.symbolic.Automaton`
    @return: node in a `dd.bdd.BDD`
    @rtype: `int`
    """
    int_types = {'int', 'saturating', 'modwrap'}
    bdd = aut.bdd
    u = bdd.true
    for var, d in aut.vars.iteritems():
        t = d['type']
        if t == 'bool':
            continue
        assert t in int_types, t
        dom = d['dom']
        p, q = dom
        e = "({p} <= {var}) & ({var} <= {q})".format(
            p=p, q=q, var=var)
        v = aut.add_expr(e)
        u = bdd.apply('and', u, v)
    return u


def _strategy_to_state_annotated(g, aut):
    """Move annotation to `dict` as value of `'state'` key.

    @type g: `nx.DiGraph`
    @type: aut: `omega.symbolic.symbolic.Automaton`
    @rtype: `nx.DiGraph`
    """
    h = nx.DiGraph()
    for u, d in g.nodes_iter(data=True):
        dvars = {k: d[k] for k in d if k in aut.vars}
        h.add_node(u, state=dvars)
    for u, v in g.edges_iter():
        h.add_edge(u, v)
    return h


def _grspec_to_automaton(g):
    """Return `symbolic.Automaton` from `GRSpec`.

    @type g: `tulip.spec.form.GRSpec`
    @rtype: `omega.symbolic.symbolic.Automaton`
    """
    if omega is None:
        raise ImportError(
            'Failed to import package `omega`.')
    a = sym.Automaton()
    d = dict(g.env_vars)
    d.update(g.sys_vars)
    for k, v in d.iteritems():
        if v in ('boolean', 'bool'):
            r = 'bool'
        elif isinstance(v, list):
            # string var -> integer var
            r = (0, len(v) - 1)
        elif isinstance(v, tuple):
            r = v
        else:
            raise ValueError(
                'unknown variable type: {v}'.format(v=v))
        d[k] = r
    g.str_to_int()
    # reverse mapping by `synth.strategy2mealy`
    a.vars = bv.make_table(d, env_vars=g.env_vars)
    f = g._bool_int.__getitem__
    a.init['env'] = map(f, g.env_init)
    a.init['sys'] = map(f, g.sys_init)
    a.action['env'] = map(f, g.env_safety)
    a.action['sys'] = map(f, g.sys_safety)
    a.win['<>[]'] = [
        '!({s})'.format(s=s)
        for s in map(f, g.env_prog)]
    a.win['[]<>'] = map(f, g.sys_prog)
    a.moore = g.moore
    a.plus_one = g.plus_one
    a.qinit = g.qinit
    return a



# Copyright (c) 2013 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""
Transition System module developer examples
"""
from scipy.sparse import lil_matrix
from numpy.random import rand
import networkx as nx
import tulip.transys as trs
import warnings

hl = 60*'='
save_fig = False

def sims_demo():
    """Storing simulations.
    """
    #=============
    # Sequences
    #=============
    # fts def:
    #   s1 -[a]-> s2
    #   s2 -[b]-> s3
    #   s3 -[a]-> s3
    prefix = ['s1', 'a', 's2', 'b', 's3']
    suffix = ['s3', 'a', 's2', 'b', 's3']
    execution = trs.InfiniteSequence(prefix, suffix)
    #print(execution)

    # trace def
    prefix = [{'p1', 'p2'}, {}, {'p1'} ]
    suffix = [{'p1'}, {}, {'p1'} ]
    trace = trs.InfiniteSequence(prefix, suffix)

    #print(trace.get_prefix() )
    #print(trace.get_suffix() )

    #=============
    # FTS Sim
    #=============
    fts_sim = trs.FiniteTransitionSystemSimulation(execution, trace)

    print('Execution of FTS: s0, a1, s1, ..., aN, sN')
    print(fts_sim.execution)

    print('Path of states: s1, s2, ..., sN')
    print(fts_sim.path)

    print('Trace of stat labels: L(s0), L(s1), ..., L(sN)')
    print(fts_sim.trace)

    print('Action Trace: a1, a2, ..., aN')
    print(fts_sim.action_trace)

    print(fts_sim)

    #=============
    # Automaton Sim
    #=============
    prefix = [{'p1'}, {'p3'}, {'p2'}]
    suffix = [{'p2'}, {'p1'}, {'p2'}]
    input_word = trs.InfiniteWord(prefix, suffix)

    prefix = ['s1', 's2', 's3']
    suffix = ['s3', 's2', 's3']
    run = trs.InfiniteSequence(prefix, suffix)

    aut_sim = trs.FiniteStateAutomatonSimulation(input_word, run)
    print(aut_sim)

def fts_maximal_example():
    """Finite-Transition System demo.
    """

    print(hl +'\nClosed FTS   -    Example 2.2, p.21 [Baier]\n' +hl)
    fts = trs.FiniteTransitionSystem()
    fts.name = 'Beverage vending machine'

    # add state info
    fts.states.add('pay')
    fts.states.remove('pay')

    fts.states.add_from({'pay', 'soda', 'select', 'beer'} )
    fts.states.remove_from({'pay', 'soda'} )
    fts.states.add_from({'pay', 'soda'} )

    fts.states.current = ['pay']

    fts.states.initial.add('pay') # should already be a state
    fts.states.initial |= {'soda', 'select'}
    fts.states.initial.remove('soda')

    fts.states.add('water')
    fts.states.initial.add('water')
    fts.states.initial -= {'water', 'select'}

    # no transitions yet...
    pre = fts.states.pre({'pay'} )
    post = fts.states.post({'pay'} )
    print("Pre('pay') = " +str(pre) )
    print("Post('pay') = " +str(post) )

    try:
        fts.states.initial.add('not pay')
    except:
        warnings.warn('You cannot add an initial state \\notin states.')

    # get state info
    print('States:\n\t' +str(fts.states() ) )
    print('Number of states:\n\t' +str(len(fts.states) ) )
    print('Initial states:\n\t' +str(fts.states.initial) )
    print('Number of initial states:\n\t' +str(len(fts.states.initial) ) )
    print('Current state:\n\t' +str(fts.states.current) )

    print("Is 'pay' a state ?\n\t" +str('pay' in fts.states) )
    print("Is 'not pay' a state ?\n\t" +str('not pay' in fts.states() ) )
    print("Is 'bla' a state ?\n\t" +str('bla' in fts) )
    print('')

    fts.plot()

    # add transition info (unlabeled)
    fts.transitions.add('pay', 'select') # notice: no labels
    fts.transitions.add_from(
        [('select', x) for x in {'soda', 'beer'}]
    )
    fts.transitions.add_from(
        [('soda', x) for x in {'beer', 'pay'}]
    )
    fts.transitions.add('pay', 'soda')
    fts.transitions.remove('pay', 'soda')

    pre = fts.states.pre({'pay'} )
    post = fts.states.post({'pay'} )
    print("Pre('pay') =\n\t" +str(pre) )
    print("Post('pay') =\n\t" +str(post) +'\n')

    # another way
    fts['pay'] # post of 'pay', note the edge ids returned, due to MultiDiGraph
    fts['select']
    fts['select']['soda']

    try:
        pre = fts.states.pre('pay')
    except:
        print('pre cannot distinguish single states from sets of states, '+
               'because a state can be anything, e.g., something iterable.\n')
    # use instead
    pre = fts.states.pre('pay')

    # all 1-hop post sets
    print('\n' +10*'-' +'\nPost sets are:\n')
    for state in fts.states():
        post = fts.states.post({state} )
        # same as:
        # post = fts.states.post_signle(state)

        print('\tof state: ' +str(state) +', the states: ' +str(post) )
    print(10*'-' +'\n')

    # same thing
    post_all = fts.states.post(fts.states() )
    print('Same thing as above:\n\t' +str(post_all) +'\n' +10*'-' +'\n')

    try:
        fts.transitions.add('pay', 'not yet a state')
    except:
        print('First add from_, to_ states, then you can add transition.\n')

    try:
        fts.transitions.add_from({'not a state', 'also not a state'}, {'pay'} )
    except:
        print('Same state check as above.\n')

    # get transition info (unlabeled)
    print('Transitions:\n\t' +str(fts.transitions() ) )
    print('Number of transitions:\n\t' +str(len(fts.transitions) ) +'\n')

    fts.plot()
    print(fts) # pretty

    # ----------------------------------------
    # CAUTION: labeling now considered
    # ----------------------------------------

    # MultiDiGraph labeling issue vs multiple same edges resolved as follows:
    #   transition addition is strictly monitored
    #   each transition is identified uniquely with its set of labels (key=values)
    #   see Transitions.add_labeled()
    fts.sys_actions.add('insert_coin')
    fts.sys_actions.add_from({'get_soda', 'get_beer', ''} )

    try:
        fts.transitions.add_labeled('pay', 'select', 'insert_coin')
    except:
        print('Checks not pre-existing same labeled or unlabeled')

    # checking pre-existing CANNOT be overriden,
    # to preserve function semantics
    # check=False used only to add missing states or labels

    # first remove unlabeled, then add new labeled
    fts.transitions.remove('pay', 'select')
    fts.plot()
    fts.transitions.add('pay', 'select', sys_actions='insert_coin')
    fts.plot()
    fts.transitions.remove('pay', 'select', sys_actions='insert_coin')

    try:
        fts.transitions.add_labeled('pay', 'new state', 'insert_coin')
    except:
        print('trying to add labeled with new state fails.\n')

    try:
        fts.transitions.add_labeled('pay', 'select', 'new action')
    except:
        print('trying to add transition with new label also fails.\n')

    # to override and add new state and/or new labels
    fts.plot()
    fts.states.add_from({'pay', 'new_state'})
    fts.sys_actions.add('new_action')

    fts.transitions.add('pay', 'new_state', sys_actions='new_action')
    fts.plot()
    fts.states.remove('new_state')
    fts.sys_actions.remove('new_action')
    fts.plot()

    fts.transitions.add('pay', 'select', sys_actions='insert_coin')
    fts.transitions.remove_from(
        [('select', x) for x in {'soda', 'beer'}]
    )
    fts.transitions.add_from(
        [('select', x) for x in {'soda', 'beer'}]
    )
    fts.transitions.remove('soda', 'pay')
    fts.transitions.add('soda', 'pay', sys_actions='get_soda')

    fts.plot()

    print('Types of actions: ' +str(fts._transition_label_def.keys() ) )
    print('Number of actions: ' +str(len(fts.sys_actions) ) )
    print('Actions: ' +str(fts.sys_actions ) )
    print('Labeled transitions: ' +str(fts.transitions() ) )
    fts.plot()

    # fast way to get all edges with value of actions
    nx.get_edge_attributes(fts, 'sys_actions')

    # Atomic Propositions (AP)
    fts.atomic_propositions.add('paid')
    fts.atomic_propositions.add_from({'', 'drink'} )

    fts.states.add('pay')

    nodes = {'soda', 'beer'}
    labels = {'paid', 'drink'}
    for node, label in zip(nodes, labels):
        fts.add_node(node, ap={label})
    fts.states.add('select', ap={'paid'})

    fts.plot()

    # export
    print(hl +'\n CAUTION: Saving DOT, PDF files\n' +hl +'\n')
    path = './test_fts'
    dot_fname = path +'.dot'
    pdf_fname = path +'.pdf'

    if not fts.plot() and save_fig:
        fts.save(pdf_fname)
        #fts.save(dot_fname, 'dot')

def ba_maximal_example():
    """Buchi Automaton demo.
    """
    print(hl +'\nBuchi Automaton\n' +hl)
    ba = trs.BuchiAutomaton(atomic_proposition_based=True)

    ba.states.add('q0')
    ba.states.add_from({'q1', 'q2', 'q3'})

    ba.states.initial.add('q0')
    ba.plot()

    # alternatives to add props
    ba.alphabet.math_set.add(['paid'] )
    ba.atomic_propositions |= {'drink', 'paid'}

    print('Number of letters: ' +str(len(ba.alphabet) ) +'\n')
    print('Alphabet: ' +str(ba.alphabet) +'\n')

    try:
        ba.transitions.add_labeled('q1', 'q10', {'paid'} )
    except:
        print('q10 not a state.')

    ba.transitions.add('q0', 'q1', letter=set())
    ba.transitions.add('q0', 'q1', letter={'paid'})
    ba.transitions.add('q1', 'q2', letter={'paid', 'drink'})
    ba.transitions.add('q3', 'q0', letter=set())
    ba.transitions.add('q1', 'q3', letter={'drink'})
    ba.plot()

    # accepting states
    ba.states.accepting.add('q1')
    ba.states.accepting |= {'q2', 'q3'}
    ba.states.accepting.remove('q2')
    ba.states.accepting.remove('q3')

    print('Number of accepting states:\n\t' +
          str(len(ba.states.accepting) ) +'\n')
    print('Accepting states:\n\t' +str(ba.states.accepting) +'\n')
    print(ba)

    path = './test_ba'
    dot_fname = path +'.dot'
    pdf_fname = path +'.pdf'

    if not ba.plot() and save_fig:
        ba.save(pdf_fname)
        #ba.save(dot_fname, 'dot')

    return ba

def scipy_sparse_labeled_adj():
    n = 10

    A = lil_matrix((n, n) )
    A[0, :3] = rand(3)
    adj2states = range(n)

    print(A)

    ofts = trs.FTS()
    ofts.states.add_from(set(range(10) ) )

    ofts.sys_actions.add('move')
    ofts.env_actions.add('rain')

    ofts.transitions.add_adj(
        A, adj2states,
        sys_actions='move',
        env_actions='rain'
    )

    ofts.plot()

    """same thing as above, using A as a submatrix instead
    """
    A = lil_matrix((3, 3) )
    A[0, :3] = rand(3)
    adj2states = [0, 1, 2]

    print(A)

    ofts = trs.FTS()
    ofts.states.add_from(set(range(10) ) )

    ofts.sys_actions.add('move')
    ofts.env_actions.add('rain')

    ofts.transitions.add_adj(
        A, adj2states,
        sys_actions='move',
        env_actions='rain'
    )
    ofts.plot()

    return ofts

def label_per_state():
    """Add states with (possibly) different AP labels each.
    """
    fts = trs.FTS()
    fts.states.add_from(['s0', 's1'] )
    fts.atomic_propositions.add('p')
    fts.states.add('s0', ap={'p'})
    fts.states.add('s1', ap=set())
    fts.plot()

if __name__ == '__main__':
    #sims_demo()
    fts_maximal_example()
    #ofts_maximal_example()
    ba_maximal_example()

    ofts = scipy_sparse_labeled_adj()
    label_per_state()
